from mot.library_functions import LibNMSimplex
from ...cl_routines.optimizing.base import AbstractParallelOptimizer, AbstractParallelOptimizerWorker

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NMSimplex(AbstractParallelOptimizer):

    default_patience = 200

    def __init__(self, scale=None, alpha=None, beta=None, gamma=None, delta=None, adaptive_scales=None,
                 patience=None, optimizer_settings=None, **kwargs):
        """Use the Nelder-Mead simplex method to calculate the optimimum.

        Args:
            patience (int): Used to set the maximum number of iterations to patience*(number_of_parameters+1)
            scale (double): the scale of the initial simplex, default 1.0
            alpha (double): reflection coefficient, default 1.0
            beta (double): contraction coefficient, default 0.5
            gamma (double); expansion coefficient, default 2.0
            delta (double); shrinkage coefficient, default 0.5
            adaptive_scales (boolean): if set to True we use adaptive scales instead of the default scale values.
                This sets the scales to:

                .. code-block:: python

                    n = model.get_nmr_estimable_parameters()

                    alpha = 1
                    beta  = 0.75 - 1.0 / (2 * n)
                    gamma = 1 + 2.0 / n
                    delta = 1 - 1.0 / n

                Following the paper:

                * Gao F, Han L. Implementing the Nelder-Mead simplex algorithm with adaptive parameters.
                  Comput Optim Appl. 2012;51(1):259-277. doi:10.1007/s10589-010-9329-3.

            optimizer_settings (dict): the optimization settings, you can use the following:
                scale, alpha, beta, gamma, delta, adaptive_scales

                The scales should satisfy the following constraints:

                .. code-block:: python

                    alpha > 0
                    0 < beta < 1
                    gamma > 1
                    gamma > alpha
                    0 < delta < 1
        """
        patience = patience or self.default_patience

        optimizer_settings = optimizer_settings or {}

        keyword_values = {}
        keyword_values['scale'] = scale
        keyword_values['alpha'] = alpha
        keyword_values['beta'] = beta
        keyword_values['gamma'] = gamma
        keyword_values['delta'] = delta
        keyword_values['adaptive_scales'] = adaptive_scales

        option_defaults = {'alpha': 1.0, 'beta': 0.5, 'gamma': 2.0, 'delta': 0.5, 'scale': 1.0,
                           'adaptive_scales': True}

        def get_value(option_name):
            value = keyword_values.get(option_name)
            if value is None:
                value = optimizer_settings.get(option_name)
            if value is None:
                value = option_defaults[option_name]
            return value

        for option in option_defaults:
            optimizer_settings.update({option: get_value(option)})

        super(NMSimplex, self).__init__(patience=patience, optimizer_settings=optimizer_settings, **kwargs)

    def minimize(self, model, init_params=None):
        if self._optimizer_settings.get('adaptive_scales', True):
            nmr_params = model.get_nmr_estimable_parameters()
            self._optimizer_settings.update(
                {'alpha': 1,
                 'beta': 0.75 - 1.0 / (2 * nmr_params),
                 'gamma': 1 + 2.0 / nmr_params,
                 'delta': 1 - 1.0 / nmr_params}
                )
        return super(NMSimplex, self).minimize(model, init_params=init_params)

    def _get_worker_generator(self, *args):
        return lambda cl_environment: NMSimplexWorker(cl_environment, *args)


class NMSimplexWorker(AbstractParallelOptimizerWorker):

    def _get_optimization_function(self):
        params = {'NMR_PARAMS': self._nmr_params, 'PATIENCE': self._parent_optimizer.patience}

        for option, value in self._optimizer_settings.items():
            if option == 'scale':
                params['INITIAL_SIMPLEX_SCALES'] = '{' + ', '.join([str(value)] * self._nmr_params) + '}'
            else:
                params.update({option.upper(): value})

        lib_nmsimplex = LibNMSimplex(evaluate_fname='evaluate')
        body = lib_nmsimplex.get_cl_code()

        body += '''
            int nmsimplex(mot_float_type* const model_parameters, void* data){
                mot_float_type initial_simplex_scale[%(NMR_PARAMS)r] = %(INITIAL_SIMPLEX_SCALES)s;
                mot_float_type fdiff;
                mot_float_type psi = 0;
                mot_float_type nmsimplex_scratch[%(NMR_PARAMS)r * 2 + (%(NMR_PARAMS)r + 1) * (%(NMR_PARAMS)r + 1)];

                return lib_nmsimplex(%(NMR_PARAMS)r, model_parameters, data, initial_simplex_scale,
                                     &fdiff, psi, (int)(%(PATIENCE)r * (%(NMR_PARAMS)r+1)),
                                     %(ALPHA)r, %(BETA)r, %(GAMMA)r, %(DELTA)r,
                                     nmsimplex_scratch);
            }
        ''' % params

        return body

    def _get_optimizer_call_name(self):
        return 'nmsimplex'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_mot
----------------------------------

Tests for `mot` module.
"""

import unittest

import numpy as np

import mot
from mot import configuration
from mot.cl_routines.optimizing.nmsimplex import NMSimplex
from mot.cl_routines.optimizing.levenberg_marquardt import LevenbergMarquardt
from mot.cl_routines.optimizing.powell import Powell
from mot.cl_routines.filters.gaussian import GaussianFilter
from mot.cl_routines.filters.mean import MeanFilter
from mot.cl_routines.filters.median import MedianFilter
from mot.utils import SimpleNamedCLFunction, convert_data_to_dtype

from mot.model_interfaces import OptimizeModelInterface


class CLRoutineTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(CLRoutineTestCase, self).__init__(*args, **kwargs)
        self._old_config_value = mot.configuration._config['compile_flags']['general']['-cl-single-precision-constant']

    def setUp(self):
        mot.configuration._config['compile_flags']['general'].update({
            '-cl-single-precision-constant': False
        })

    def tearDown(self):
        mot.configuration._config['compile_flags']['general'].update({
            '-cl-single-precision-constant': self._old_config_value
        })


class TestRosenbrock(CLRoutineTestCase):

    def setUp(self):
        super(TestRosenbrock, self).setUp()
        self.model = Rosenbrock(5)
        self.optimizers = (NMSimplex(), Powell(patience=10))

    def test_model(self):
        for optimizer in self.optimizers:
            v = optimizer.minimize(self.model).get_optimization_result()[0]
            for ind in range(self.model.get_nmr_inst_per_problem()):
                self.assertAlmostEqual(float(v[ind]), 1.0, places=3)


class TestLSQNonLinExample(CLRoutineTestCase):

    def setUp(self):
        super(TestLSQNonLinExample, self).setUp()
        self.model = MatlabLSQNonlinExample()
        self.optimizers = (LevenbergMarquardt(),)
        self.residual_calc = ResidualCalculator()

    def test_model(self):
        for optimizer in self.optimizers:
            v = optimizer.minimize(self.model).get_optimization_result()
            res = self.residual_calc.calculate(self.model, v)
            s = 0
            for i in range(res.shape[1]):
                s += res[0, i]**2
            self.assertAlmostEqual(s, 124.3622, places=4)


class TestFilters(CLRoutineTestCase):

    def setUp(self):
        super(TestFilters, self).setUp()
        self.d1 = np.array([1, 2, 4, 2, 1], dtype=np.float64)
        self.d2 = np.eye(4)

    def test_median(self):
        filter = MedianFilter(2)
        s1 = filter.filter(self.d1)
        np.testing.assert_almost_equal(s1, np.array([2, 2, 2, 2, 2]))

        s2 = filter.filter(self.d2)
        np.testing.assert_almost_equal(s2, np.zeros((4, 4)))

    def test_mean(self):
        filter = MeanFilter(2)
        s1 = filter.filter(self.d1)
        np.testing.assert_almost_equal(s1, np.array([2 + 1/3.0, 2.25, 2, 2.25, 2 + 1/3.0]))

        s2 = filter.filter(self.d2)
        expected = np.ones((4, 4)) * 0.25
        expected[0, 0] = 1/3.0
        expected[0, 3] = 2/9.0
        expected[3, 0] = 2/9.0
        expected[3, 3] = 1/3.0
        np.testing.assert_almost_equal(s2, expected)

    def test_gaussian(self):
        filter = GaussianFilter(2, sigma=1.0)
        s1 = filter.filter(self.d1, mask=np.array([1, 1, 1, 1, 0]))
        s2 = filter.filter(self.d2)

        np.testing.assert_almost_equal(s1, [1.1089774, 2.135224, 2.6417738, 1.8910226, 0])

        expected = np.array([[0.22470613, 0.20994687, 0.10351076, 0.02661242],
                             [0.20994687, 0.28434043, 0.22325308, 0.10351076],
                             [0.10351076, 0.22325308, 0.28434043, 0.20994687],
                             [0.02661242, 0.10351076, 0.20994687, 0.22470613]])
        np.testing.assert_almost_equal(s2, expected)



class Rosenbrock(OptimizeModelInterface):

    def __init__(self, n=5):
        """When optimized the parameters should all be equal to 1."""
        super(OptimizeModelInterface, self).__init__()
        self.n = n

    def double_precision(self):
        return True

    @property
    def name(self):
        return 'rosenbrock'

    def get_kernel_data(self):
        return []

    def get_nmr_problems(self):
        return 1

    def get_pre_eval_parameter_modifier(self):
        func_name = '_modifyParameters'
        func = '''
            void ''' + func_name + '''(void* data, mot_float_type* x){
            }
        '''
        return SimpleNamedCLFunction(func, func_name)

    def get_model_eval_function(self):
        fname = 'evaluateModel'
        func = '''
            double ''' + fname + '''(void* data, const double* const x, uint observation_index){
                double sum = 0;
                for(uint i = 0; i < ''' + str(self.n) + ''' - 1; i++){
                    sum += 100 * pown((x[i + 1] - pown(x[i], 2)), 2) + pown((x[i] - 1), 2);
                }
                return -sum;
            }
        '''
        return SimpleNamedCLFunction(func, fname)

    def _get_observation_return_function(self):
        fname = 'getObservation'
        func = '''
            double ''' + fname + '''(void* data, const uint observation_index){
                return 0;
            }
        '''
        return SimpleNamedCLFunction(func, fname)

    def get_objective_per_observation_function(self):
        eval_func = self.get_model_eval_function()
        obs_func = self._get_observation_return_function()

        func = eval_func.get_cl_code()
        func += obs_func.get_cl_code()

        func_name = "getObjectiveInstanceValue"
        func += '''
            mot_float_type ''' + func_name + '''(void* data, const mot_float_type* const x, uint observation_index){
                return ''' + obs_func.get_cl_function_name() + '''(data, observation_index) -
                            ''' + eval_func.get_cl_function_name() + '''(data, x, observation_index);
            }
        '''
        return SimpleNamedCLFunction(func, func_name)

    def get_initial_parameters(self):
        params = np.ones((1, self.n)) * 3
        return convert_data_to_dtype(params, 'double')

    def get_lower_bounds(self):
        return [-np.inf] * self.n

    def get_upper_bounds(self):
        return [np.inf] * self.n

    def get_free_param_names(self):
        return list(map(str, range(self.n)))

    def get_nmr_inst_per_problem(self):
        return 1

    def get_nmr_estimable_parameters(self):
        return self.n

    def finalize_optimized_parameters(self, parameters):
        return parameters

class MatlabLSQNonlinExample(OptimizeModelInterface):

    def __init__(self):
        """When optimized the parameters should be close to [0.2578, 0.2578] or something with a similar 2 norm.

        See the matlab manual page at http://nl.mathworks.com/help/optim/ug/lsqnonlin.html for more information.
        (viewed at 2015-04-02).

        """
        super(OptimizeModelInterface, self).__init__()

    def double_precision(self):
        return True

    @property
    def name(self):
        return 'matlab_lsqnonlin_example'

    def get_kernel_data(self):
        return []

    def get_nmr_problems(self):
        return 1

    def get_pre_eval_parameter_modifier(self):
        func_name = '_modifyParameters'
        func = '''
            void ''' + func_name + '''(void* data, mot_float_type* x){
            }
        '''
        return SimpleNamedCLFunction(func, func_name)

    def get_model_eval_function(self):
        fname = 'evaluateModel'
        func = '''
            double ''' + fname + '''(void* data, const double* const x, uint k){
                return -(2 + 2 * (k+1) - exp((k+1) * x[0]) - exp((k+1) * x[1]));
            }
        '''
        return SimpleNamedCLFunction(func, fname)

    def _get_observation_return_function(self):
        fname = 'getObservation'
        func = '''
            double ''' + fname + '''(void* data, uint observation_index){
                return 0;
            }
        '''
        return SimpleNamedCLFunction(func, fname)

    def get_objective_per_observation_function(self):
        eval_func = self.get_model_eval_function()
        obs_func = self._get_observation_return_function()

        func = eval_func.get_cl_code()
        func += obs_func.get_cl_code()

        func_name = "getObjectiveInstanceValue"

        func += '''
            mot_float_type ''' + func_name + '''(void* data, const mot_float_type* const x, uint observation_index){
                return ''' + obs_func.get_cl_function_name() + '''(data, observation_index) -
                            ''' + eval_func.get_cl_function_name() + '''(data, x, observation_index);
            }
        '''
        return SimpleNamedCLFunction(func, func_name)

    def get_initial_parameters(self):
        params = np.array([[0.3, 0.4]])
        return convert_data_to_dtype(params, 'double')

    def get_lower_bounds(self):
        return [0, 0]

    def get_upper_bounds(self):
        return [np.inf] * 2

    def get_free_param_names(self):
        return ['0', '1']

    def get_nmr_inst_per_problem(self):
        return 10

    def get_nmr_estimable_parameters(self):
        return 2

    def finalize_optimized_parameters(self, parameters):
        return parameters

if __name__ == '__main__':
    unittest.main()



__author__ = 'Robbert Harms'
__date__ = "2017-03-02"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ProposalUpdate(object):

    def get_update_function(self, proposal_parameters, address_space='private'):
        """Get the update function to update the proposal parameters of one of the proposals.

        Args:
            proposal_parameters (list of ProposalParameter): the list of proposal parameters to (possibly use) in the
                update function. It will only use the parameters that have ``adaptable`` set to True.
            address_space (str): the address space to use for the function parameters
        """
        raise NotImplementedError()

    def get_function_name(self, proposal_parameters):
        """Get the name of the proposal update function.

        Args:
            proposal_parameters (list of ProposalParameter): the list of proposal parameters to (possibly use) in the
                update function. It will only use the parameters that have ``adaptable`` set to True.

        Returns:
            str: the name of the function returned by :meth:`get_update_function`
        """
        raise NotImplementedError()

    def uses_parameter_variance(self):
        """Check if this proposal update function uses the parameter variance.

        If not, we will not provide it, this saves memory in the kernel.

        Returns:
            boolean: if this proposal update function uses the parameter variance
        """
        raise NotImplementedError()

    def uses_jump_counters(self):
        """Check if this proposal update function uses the jump counters (jump counter and acceptance counter).

        If not, we will not provide it, this saves memory in the kernel.

        Returns:
            boolean: if this proposal update function uses the jump counters
        """
        raise NotImplementedError()


class SimpleProposalUpdate(ProposalUpdate):

    def __init__(self, function_name, uses_parameter_variance=False, uses_jump_counters=True):
        """A simple proposal update function template.

        Args:
            function_name (str): the name of this proposal update function, try to choose an unique name.
            uses_parameter_variance (boolean): if this proposal requires the parameter variance. Enable if you need it
                 in your update function.
            uses_jump_counters (boolean): if this proposal uses jump counters for its workings. This is enabled
                by default. You can disable it for speed.
        """
        self._function_name = function_name
        self._uses_parameter_variance = uses_parameter_variance
        self._uses_jump_counters = uses_jump_counters

    def get_update_function(self, proposal_parameters, address_space='private'):
        return self._update_function_template('', proposal_parameters, address_space)

    def _update_function_template(self, function_body, proposal_parameters, address_space):
        params = ['{address_space} mot_float_type* const {name}'.format(address_space=address_space, name=p.name)
                  for p in proposal_parameters if p.adaptable]

        if self.uses_jump_counters():
            params.extend(['{address_space} ulong* const sampling_counter'.format(address_space=address_space),
                           '{address_space} ulong* const acceptance_counter'.format(address_space=address_space)])

        if self.uses_parameter_variance():
            params.append('mot_float_type parameter_variance'.format(address_space=address_space))

        return '''
            #ifndef {include_guard_name}
            #define {include_guard_name}

            void {function_name}({params}){{
                {function_body}
            }}

            #endif //{include_guard_name}

        '''.format(include_guard_name=self.get_function_name(proposal_parameters).upper(),
                   function_name=self.get_function_name(proposal_parameters), params=', '.join(params),
                   function_body=function_body)

    def get_function_name(self, proposal_parameters):
        return 'proposal_update_{}_{}'.format(self._function_name,
                                              len([True for p in proposal_parameters if p.adaptable]))

    def uses_parameter_variance(self):
        return self._uses_parameter_variance

    def uses_jump_counters(self):
        return self._uses_jump_counters


class NoOperationUpdateFunction(SimpleProposalUpdate):

    def __init__(self):
        """This is the no-operation update function. It does not update the proposal parameters."""
        super(NoOperationUpdateFunction, self).__init__('no_opt')

    def get_update_function(self, proposal_parameters, address_space='private'):
        return self._update_function_template('', proposal_parameters, address_space)


class AcceptanceRateScaling(SimpleProposalUpdate):

    def __init__(self, target_acceptance_rate=0.44, batch_size=50, damping_factor=1, min_val=1e-15, max_val=1e3):
        """Scales the proposal parameter (typically the std) such that it oscillates towards the chosen acceptance rate.

        This uses an scaling similar to the 'Adaptive Metropolis-Within-Gibbs' method described in:
        "Examples of Adaptive MCMC", Gareth O. Roberts & Jeffrey S. Rosenthal (2009)

        This class implements the delta function as: :math:`\delta(n) = \sqrt{1 / (d*n)}`.
        Where n is the current batch index and d is the damping factor.

        As an example, with a damping factor of 500, delta reaches a scaling of 0.01 in 20 batches. At a batch size
        of 50 that would amount to 1000 samples.

        Args:
            target_acceptance_rate (float): the target acceptance rate between 0 and 1.
            batch_size (int): the size of the batches inbetween which we update the parameters
            min_val (float): the minimum value the standard deviation can take
            max_val (float): the maximum value the standard deviation can take
        """
        super(AcceptanceRateScaling, self).__init__('acceptance_rate_scaling')
        self._target_acceptance_rate = target_acceptance_rate
        self._batch_size = batch_size
        self._damping_factor = damping_factor
        self._min_val = min_val
        self._max_val = max_val

        if target_acceptance_rate > 1 or target_acceptance_rate < 0:
            raise ValueError('The target acceptance rate should be '
                             'within [0, 1], {} given.'.format(target_acceptance_rate))

    def get_update_function(self, proposal_parameters, address_space='private'):
        body = '''
            if(*sampling_counter % {batch_size} == 0){{

                mot_float_type delta = sqrt(1.0/({damping_factor} * (*sampling_counter / {batch_size})));

                if(*acceptance_counter / (mot_float_type){batch_size} > {target_ar}){{
                    *std *= exp(delta);
                }}
                else{{
                    *std /= exp(delta);
                }}

                *std = clamp(*std, (mot_float_type){min_val}, (mot_float_type){max_val});

                *acceptance_counter = 0;
            }}
        '''.format(batch_size=self._batch_size, target_ar=self._target_acceptance_rate,
                   damping_factor=self._damping_factor, min_val=self._min_val, max_val=self._max_val)
        return self._update_function_template(body, proposal_parameters, address_space)


class FSLAcceptanceRateScaling(SimpleProposalUpdate):

    def __init__(self, batch_size=50, min_val=1e-15, max_val=1e3):
        """An acceptance rate scaling algorithm found in a Neuroscience package called FSL.

        This scaling algorithm scales the std. by :math:`\sqrt(a/(n - a))` where a is the number of accepted samples
        in the last batch and n is the batch size. Its goal is to balance the acceptance rate at 0.5.

        So far, the author of this function in MOT has not been able to find theoretical support for this scaling
        algorithm. Please use this heuristic with caution.

        To prevent runaway proposal values we clamp the updated parameter value between a minimum and maximum value
        specified in the constructor.

        Args:
            batch_size (int): the size of the batches in between which we update the parameters
            min_val (float): the minimum value the standard deviation can take
            max_val (float): the maximum value the standard deviation can take
        """
        super(FSLAcceptanceRateScaling, self).__init__('fsl_acceptance_rate_scaling')
        self._batch_size = batch_size
        self._min_val = min_val
        self._max_val = max_val

    def get_update_function(self, proposal_parameters, address_space='private'):
        body = '''
            if(*sampling_counter == {batch_size}){{
                *std = clamp(*std * sqrt((mot_float_type)(*acceptance_counter + 1) /
                                         ({batch_size} - *acceptance_counter + 1)),
                             (mot_float_type){min_val},
                             (mot_float_type){max_val});

                *sampling_counter = 0;
                *acceptance_counter = 0;
            }}
        '''.format(batch_size=self._batch_size, min_val=self._min_val, max_val=self._max_val)
        return self._update_function_template(body, proposal_parameters, address_space)


class SingleComponentAdaptiveMetropolis(SimpleProposalUpdate):

    def __init__(self, waiting_period=100, scaling_factor=2.4, epsilon=1e-20):
        """Uses the Single Component Adaptive Metropolis (SCAM) scheme to update the proposals.

        This uses an scaling described in: "Componentwise adaptation for high dimensional MCMC",
        Heikki Haario, Eero Saksman and Johanna Tamminen (2005). That is, it updates the proposal standard deviation
        using the variance of the chain's history.

        Args:
            waiting_period (int): only start updating the proposal std. after this many draws.
            scaling_factor (float): the scaling factor to use (the parameter ``s`` in the paper referenced).
            epsilon (float): small number to prevent the std. from collapsing to zero.
        """
        super(SingleComponentAdaptiveMetropolis, self).__init__('scam', uses_parameter_variance=True)
        self._waiting_period = waiting_period
        self._scaling_factor = scaling_factor
        self._epsilon = epsilon

    def get_update_function(self, proposal_parameters, address_space='private'):
        body = '''
            if(*sampling_counter > {waiting_period}){{
                *std = {scaling_factor} * sqrt(parameter_variance) + {epsilon};
            }}
        '''.format(waiting_period=self._waiting_period, scaling_factor=self._scaling_factor, epsilon=self._epsilon)
        return self._update_function_template(body, proposal_parameters, address_space)



"""The interfaces needed for models.

Since a lot of information about a model is needed to be able to optimize or sample it, we encapsulate all that
information in an interface. Only objects that successful implement the interfaces in this module can be optimized or
sampled using one of the optimization or sampling routines in MOT.

These interfaces expose data and modeling code. The data is represented as numpy arrays and the CL code as strings.
"""

__author__ = 'Robbert Harms'
__date__ = "2014-03-14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class OptimizeModelInterface(object):

    @property
    def name(self):
        """Get the name of this model. This should be overwritten by the implementing model.

        Returns:
            str: A string with the name of this model.
        """
        raise NotImplementedError()

    @property
    def double_precision(self):
        """Flag to signal if we should use the double float type during calculations.

        By default we ask the cl routines to use the single precision float type, you can overwrite this with
        your own flags.

        Returns:
            boolean: if we would like to use double precision floating point during the calculations
        """
        raise NotImplementedError()

    def get_kernel_data(self):
        """Return a list of data arrays we need to load into the kernel.

        Returns:
            dict[str: mot.utils.KernelInputData]: the list of data objects we need to load in the kernel
        """
        raise NotImplementedError()

    def get_nmr_problems(self):
        """Get the number of problems we need to analyze.

        Returns:
            int: A single integer specifying the number of problem instances
        """
        raise NotImplementedError()

    def get_nmr_inst_per_problem(self):
        """Get the number of instances/data points per problem.

        The minimum is one instance per problem.

        This number represents the number of data points

        Returns:
            int: the number of instances per problem.
        """
        raise NotImplementedError()

    def get_nmr_estimable_parameters(self):
        """Get the number of estimable parameters.

        Returns:
            int: the number of estimable parameters
        """
        raise NotImplementedError()

    def get_pre_eval_parameter_modifier(self):
        """Return code that needs to be run prior to model evaluation or objective function calculation.

        This is meant to contain possible parameter transformations that need to be executed only once for a
        given set of parameters. Having this in a separate function gives a speed gain.

        Model optimization routines need to be aware that they need to call this function prior to calling any of:

        * :meth:`~get_model_eval_function`
        * :meth:`~get_objective_per_observation_function`


        Returns:
            mot.utils.NamedCLFunction: a named CL function with the following signature:

                .. code-block:: c

                    void <func_name>(mot_data_struct* data, mot_float_type* x);

                Changes may happen in place in the ``x`` parameter.
        """
        raise NotImplementedError()

    def get_model_eval_function(self):
        """Get the evaluation function that evaluates the model at the given parameters.

        The returned function should not do any error calculations,
        it should merely return the result of evaluating the model for the given parameters.

        Returns:
            mot.utils.NamedCLFunction: a named CL function with the following signature:

                .. code-block:: c

                    double <func_name>(mot_data_struct* data, const mot_float_type* const x, uint observation_index);
        """
        raise NotImplementedError()

    def get_objective_per_observation_function(self):
        """Get the objective function that returns the objective value at a measurement instance.

        This should return the objective values (of each instance point) as such that when linearly summed, we have the
        complete objective value.

        Returns:
            mot.utils.NamedCLFunction: A CL function with signature:

                .. code-block:: c

                    double <func_name>(mot_data_struct* data, const mot_float_type* const x, uint observation_index);
        """
        raise NotImplementedError()

    def get_initial_parameters(self):
        """Get a two dimensional matrix with the initial parameters (starting points) for every voxel.

        Returns:
            ndarray: A two dimensional matrix with on the first axis the problem instances and on the second
                the parameter values per problem instance
        """
        raise NotImplementedError()

    def get_lower_bounds(self):
        """Get for each estimable parameter the lower bounds.

        Returns:
            list: For every estimable parameter a scalar or vector with the the lower bound(s) for that parameter.
                For infinity use np.inf.
        """
        raise NotImplementedError()

    def get_upper_bounds(self):
        """Get for each estimable parameter the upper bounds.

        Returns:
            list: For every estimable parameter a scalar or vector with the the upper bound(s) for that parameter.
                For infinity use np.inf.
        """
        raise NotImplementedError()

    def finalize_optimized_parameters(self, parameters):
        """Called by the optimization routine after optimization to finalize the optimized parameters.

        This can be used to do some automatic post-processing similar to for example what the
        ``get_pre_eval_parameter_modifier`` can do within the kernel.

        Args:
            parameters (ndarray): the set of parameters after optimization. This method may change this array
                in place.

        Returns:
            ndarray: the updated parameters. While changes may be done in place, one must return the parameters
                one would like to use.
        """
        raise NotImplementedError()


class SampleModelInterface(OptimizeModelInterface):
    """Extends the OptimizeModelInterface with information for sampling purposes.

    This specific interface is tied to sampling with the Metropolis Hastings Random Walk sampler as
    implement in :class:`mot.cl_routines.sampling.metropolis_hastings.MetropolisHastings`.

    To be able to sample a model we (in principle) need to have:

    * a log likelihood function;
    * a proposal function;
    * and a prior function


    Proposal functions can be symmetric (if it holds that ``q(x|x') == q(x'|x)``) or
    non symmetric (i.e. ``q(x|x') != q(x'|x)``). In the case of non-symmetric proposals we need to
    have a function to get the probability log likelihood of the proposal.
    This indicates the need for two more pieces of information:

    * test if the proposal is symmetric
    * proposal log PDF function


    A trick in sampling is to have auto-adapting proposals. These proposals commonly have a distribution with a
    standard deviation that varies in time. The idea is that if the distribution is too tight (low std) only a few
    of the proposed samples are accepted and we need to broaden the distribution (increase the std). On the other
    hand, if the std is too high the jumps might not get accepted. This leads us to the following
    additional functionality:

    * proposal state update function


    Since OpenCL < 2.1 does not allow state variables in functions and also does not support classes, we need to find
    a way to store the state of the proposal distribution inside the kernel function. For that, each proposal
    CL function has as additional parameter the ``proposal_state``. The initial state can be obtained from
    this class and needs to be handed to the proposal functions in the kernels.

    Finally, this interface requires to you specify a :class:`mot.cl_routines.sampling.metropolis_hastings.MHState`
    that specifies the current state of the sampler. This can be set to a default state when starting sampling
    or to the output of a previous run to continue sampling.
    """

    def get_proposal_state(self):
        """Get for every problem instance the list of parameter values to use in the the adaptable proposal.

        Returns:
            ndarray: per problem instance the proposal parameter values that are adaptable.
        """
        raise NotImplementedError()

    def get_log_likelihood_per_observation_function(self):
        """Get the (complete) CL Log Likelihood function that evaluates the given instance under a noise model.

        This should return the LL's such that when linearly summed they yield the total log likelihood of the model.

        Returns:
            mot.utils.NamedCLFunction: A function of the kind:
                .. code-block:: c

                    double <fname>(mot_data_struct* data, const mot_float_type* const x, uint observation_index);
        """
        raise NotImplementedError()

    def is_proposal_symmetric(self):
        """Check if the entire proposal distribution is symmetric: ``q(x|x') == q(x'|x)``.

        Returns:
            boolean: True if the proposal distribution is symmetric, false otherwise.
        """
        raise NotImplementedError()

    def get_proposal_logpdf(self, address_space_proposal_state='private'):
        """Get the probability density function of the proposal in log space (as a CL string).

        This density function is used if the proposal is not symmetric.

        Args:
            address_space_proposal_state (str): the CL address space of the proposal state vector.
                Defaults to ``private``.

        Returns:
            mot.utils.NamedCLFunction: A function with the signature:

                .. code-block:: c

                    double <func_name>(uint param_ind, mot_float_type proposal,
                                       mot_float_type current,
                                       <address_space_proposal_state> mot_float_type* const proposal_state);


            Where ``param_ind`` is the index of the parameter we would like to get the proposal from,
            ``current`` is the current value of that parameter and ``proposal`` the proposal value of the parameter.
            The final argument ``proposal_state`` are the current settings of the proposal function.

            It should return for the requested parameter a value ``q(proposal | current)``, the log Probability
            Density Function (log PDF) of the proposal given the current value.
        """
        raise NotImplementedError()

    def get_proposal_function(self, address_space_proposal_state='private'):
        """Get a proposal function that returns proposals for a requested parameter.

        Args:
            address_space_proposal_state (str): the CL address space of the proposal state vector.
                Defaults to ``private``.

        Returns:
            mot.utils.NamedCLFunction: A function with the signature:

                .. code-block:: c

                    mot_float_type <func_name>(
                        uint param_ind,
                        mot_float_type current,
                        void* rng_data,
                        <address_space_proposal_state> mot_float_type* const proposal_state);


            Where ``param_ind`` is the index of the parameter for which we want the proposal and
            ``current`` is the current value of that parameter. The argument ``proposal_state`` is the
            state of the proposal distribution. One can obtain random numbers with:
            .. code-block:: c

                float randomnr = frand(rng_data);
        """
        raise NotImplementedError()

    def get_proposal_state_update_function(self, address_space='private'):
        """Get the function to update the proposal parameters

        Args:
            address_space (str): the address space of (all) the given arguments, defaults to ``private``

        Returns:
            mot.utils.NamedCLFunction: A function with the signature:
                .. code-block:: c

                    void <func_name>(<address_space> mot_float_type* const proposal_state,
                                     <address_space> ulong* const sampling_counter,
                                     <address_space> ulong* const acceptance_counter);

                The ``proposal_state`` holds the current value of all the adaptable proposal parameters and is
                of length equal to the number of adaptable parameters. The ``sampling_counter`` holds the number of
                samples drawn since last update (per parameter) and ``acceptance_counter`` holds the number of samples
                that where accepted since the last update. Both are of length equal to the total number of parameters
                in the model (!). The implementing function is free to overwrite the values in each array.
        """
        raise NotImplementedError()

    def proposal_state_update_uses_variance(self):
        """Check if the proposal state update function requires the variance of each of the parameters.

        If none of the proposal update functions require the parameter variance then we can save memory in the
        kernel by not calculating them.

        Returns:
            boolean: if at least one parameter proposal state update function requires the parameter variance
                return True, else return False.
        """
        raise NotImplementedError()

    def get_log_prior_function(self, address_space_parameter_vector='private'):
        """Get the prior function that returns the prior information about the given parameters.

        The prior function must be in log space.

        Args:
            address_space_parameter_vector (str): the address space to use for the parameter vector
                by default this is set to ``private``.

        Returns:
            mot.utils.NamedCLFunction: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(
                        mot_data_struct* data,
                        <address_space_parameter_vector> const mot_float_type* const x
                    );

            Which is called by the sampling routine to calculate the posterior probability.
        """
        raise NotImplementedError()

    def get_metropolis_hastings_state(self):
        """Get the current state of the Metropolis Hastings sampler.

        This can be used to continue execution of an MH sampling from a previous point in time.

        Returns:
            mot.cl_routines.sampling.metropolis_hastings.MHState: the current Metropolis Hastings state
        """
        raise NotImplementedError()


class NumericalDerivativeInterface(OptimizeModelInterface):
    """Extends the model with information necessary for calculating numerical derivatives of the objective function.

    For calculating derivatives (gradients / Hessians) numerically, we need a likelihood function and some additional
    information, like the step size for each parameter, a method for checking boundary conditions and possible parameter
    transformations for circular parameters. All these extra elements are represented in this interface.
    """

    def numdiff_get_max_step(self):
        """Get for each estimable parameter the maximum step size to use for calculating numerical derivatives.

        The derivative calculation method typically uses an adaptive step size to determine the step with the best
        trade-off between numerical errors and localization of the derivative. This method must return the
        initial and largest step size to use.

        Returns:
            list[float]: per parameter a single float with the step size for that parameter
        """
        raise NotImplementedError()

    def numdiff_get_scaling_factors(self):
        """Get for each estimable parameter a scaling factor that is to be used for scaling this parameter to unitary.

        Since numerical differentiation is sensitive to differences in step sizes, it is better to rescale
        the parameters to a unitary range instead of changing the step sizes for the parameters.

        This should return numbers such that when the parameter is multiplied with this value, the magnitude of the
        parameter is about one.

        Returns:
            list[float]: per estimable parameter a single float with the parameter scaling to use for that parameter.
                Use 1 as identity.
        """
        raise NotImplementedError()

    def numdiff_use_bounds(self):
        """Check for each parameter if we should be using the bounds for that parameter when taking the derivative.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_use_upper_bounds(self):
        """Check for each parameter if we should be using the upper bounds when taking the derivative.

        This is only used if use_bounds is True for a parameter.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the upper bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_use_lower_bounds(self):
        """Check for each parameter if we should be using the lower bounds when taking the derivative.

        This is only used if use_bounds is True for a parameter.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the lower bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_parameter_transformation(self):
        """A transformation that can prepare the parameter +/- the proposed step for evaluation in the model function.

        Some parameters require for example a modulus operation before the proposed step can be used in the model
        This is the place to define them. Please note that this function does not need to incorporate boundary checks
        as those are handled already by the numerical differentiation routine.

        Returns:
            mot.utils.NamedCLFunction: A function with the signature:
                .. code-block:: c

                    void <func_name>(mot_data_struct* data, mot_float_type* params);

                Where the data is the kernel data struct and params is the vector with the suggested parameters and
                which can be modified in place. Note that this is called two times, one with the parameters plus
                the step and one time without.
        """
        raise NotImplementedError()



import numpy as np
from mot.cl_data_type import SimpleCLDataType
from mot.cl_function import CLFunction, SimpleCLFunction, CLPrototype, SimpleCLPrototype
from mot.model_building.parameter_functions.numdiff_info import SimpleNumDiffInfo
from mot.model_building.parameters import FreeParameter
from mot.model_building.parameter_functions.priors import ARDGaussian, UniformWithinBoundsPrior, ARDBeta
from mot.model_building.parameter_functions.proposals import GaussianProposal
from mot.model_building.parameter_functions.transformations import ClampTransform, CosSqrClampTransform

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ModelCLPrototype(CLPrototype):

    @property
    def name(self):
        """Get the name of this model function.

        Returns:
            str: The name of this model function.
        """
        raise NotImplementedError()

    def get_free_parameters(self):
        """Get all the free parameters in this model

        Returns:
            list of CLFunctionParameter: list of all the model parameters of type FreeParameter in this model
        """
        raise NotImplementedError()


class SampleModelCLPrototype(ModelCLPrototype):

    def get_prior_parameters(self, parameter):
        """Get the prior parameters of the given parameter.

        Args:
            parameter (FreeParameter): one of the parameters of this model function

        Returns:
            list of parameters: the list of prior parameters for the given parameter
        """
        raise NotImplementedError()

    def get_model_function_priors(self):
        """Get all the model function priors.

        Returns:
            list[mot.cl_function.CLFunction]: the priors for this model function,
                these are supposed to be used in conjunction to the parameter priors.
        """
        raise NotImplementedError()


class ModelCLFunction(CLFunction, ModelCLPrototype):
    """Interface for a basic model function just for optimization purposes.

    If you need to sample the model, use the extended version of this interface :class:`SampleModelCLFunction`.
    """


class SampleModelCLFunction(ModelCLFunction, SampleModelCLPrototype):
    """Extended version of a model function for use in sampling.

    This adds functions to retrieve priors about this function.
    """


class SimpleSampleModelCLPrototype(SimpleCLPrototype, SampleModelCLPrototype):

    def __init__(self, return_type, name, cl_function_name, parameter_list, model_function_priors=None):
        """This CL function is for all estimable models

        Args:
            return_type (str): the CL return type of the function
            name (str): The name of the model
            cl_function_name (string): The name of the CL function
            parameter_list (list or tuple of CLFunctionParameter): The list of parameters required for this function
            model_function_priors (list of mot.cl_function.CLFunction): list of priors concerning this whole model
                function. The parameter names of the given functions must match those of this function.
        """
        super(SimpleSampleModelCLPrototype, self).__init__(return_type, cl_function_name, parameter_list)
        self._name = name
        self._model_function_priors = model_function_priors or []
        if isinstance(self._model_function_priors, CLFunction):
            self._model_function_priors = [self._model_function_priors]

    @property
    def name(self):
        """Get the name of this model function.

        Returns:
            str: The name of this model function.
        """
        return self._name

    def get_model_function_priors(self):
        """Get all the model function priors.

        Returns:
            list[mot.cl_function.CLFunction]: the priors for this model function, these are supposed to be used in
                conjunction to the parameter priors.
        """
        return self._model_function_priors

    def get_free_parameters(self):
        """Get all the free parameters in this model

        Returns:
            list: the list of free parameters in this model
        """
        return list([p for p in self.get_parameters() if isinstance(p, FreeParameter)])

    def get_prior_parameters(self, parameter):
        """Get the parameters referred to by the priors of the free parameters.

        This returns a list of all the parameters referenced by the prior parameters, recursively.

        Returns:
            list of parameters: the list of additional parameters in the prior for the given parameter
        """
        def get_prior_parameters(params):
            return_params = []

            for param in params:
                prior_params = param.sampling_prior.get_parameters()[3:]
                proxy_prior_params = [prior_param.get_renamed('{}.prior.{}'.format(param.name, prior_param.name))
                                      for prior_param in prior_params]

                return_params.extend(proxy_prior_params)

                free_prior_params = [p for p in proxy_prior_params if isinstance(p, FreeParameter)]
                return_params.extend(get_prior_parameters(free_prior_params))

            return return_params

        return get_prior_parameters([parameter])


class SimpleModelCLFunction(SampleModelCLFunction, SimpleCLFunction):

    def __init__(self, return_type, name, cl_function_name, parameter_list, cl_body, dependency_list=(),
                 model_function_priors=None, cl_extra=None):
        """This CL function is for all estimable models

        Args:
            return_type (str): the CL return type of the function
            name (str): The name of the model
            cl_function_name (string): The name of the CL function
            parameter_list (list or tuple of CLFunctionParameter): The list of parameters required for this function
            cl_body (str): the cl body of this function
            dependency_list (list or tuple of CLFunction): The list of CL libraries this function depends on
            model_function_priors (list of mot.cl_function.CLFunction): list of priors concerning this whole model
                function. The parameter names of the given functions must match those of this function.
        """
        super(SimpleModelCLFunction, self).__init__(return_type, cl_function_name, parameter_list,
                                                    cl_body, dependency_list=dependency_list, cl_extra=cl_extra)

        self._prototype = SimpleSampleModelCLPrototype(return_type, name, cl_function_name,
                                                       parameter_list, model_function_priors=model_function_priors)

    @property
    def name(self):
        return self._prototype.name

    def get_model_function_priors(self):
        return self._prototype.get_model_function_priors()

    def get_free_parameters(self):
        return self._prototype.get_free_parameters()

    def get_prior_parameters(self, parameter):
        return self._prototype.get_prior_parameters(parameter)


class Scalar(SimpleModelCLFunction):

    def __init__(self, name='Scalar', param_name='s', value=0.0, lower_bound=0.0, upper_bound=np.inf,
                 parameter_kwargs=None):
        """A Scalar model function to be used during optimization.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
            parameter_kwargs (dict): additional settings for the parameter initialization
        """
        parameter_settings = dict(parameter_transform=ClampTransform(),
                                  sampling_proposal=GaussianProposal(1.0))
        parameter_settings.update(parameter_kwargs or {})

        super(Scalar, self).__init__(
            'mot_float_type',
            name,
            'Scalar',
            (FreeParameter(SimpleCLDataType.from_string('mot_float_type'), param_name,
                           False, value, lower_bound, upper_bound, **parameter_settings),),
            'return ' + param_name + ';')


class Weight(Scalar):

    def __init__(self, name='Weight', value=0.5, lower_bound=0.0, upper_bound=1.0, parameter_kwargs=None):
        """Implements Scalar model function to add the semantics of representing a Weight.

        Some of the code checks for type Weight, be sure to use this model function if you want to represent a Weight.

        A weight is meant to be a model volume fraction.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        parameter_settings = dict(parameter_transform=CosSqrClampTransform(),
                                  sampling_proposal=GaussianProposal(0.01),
                                  sampling_prior=UniformWithinBoundsPrior(),
                                  numdiff_info=SimpleNumDiffInfo(scale_factor=10)
                                  )
        parameter_settings.update(parameter_kwargs or {})

        super(Weight, self).__init__(name=name, param_name='w', value=value, lower_bound=lower_bound,
                                     upper_bound=upper_bound, parameter_kwargs=parameter_settings)


class ARD_Beta_Weight(Weight):

    def __init__(self, name='ARD_Beta_Weight', value=0.5, lower_bound=0.0, upper_bound=1.0):
        """A compartment weight with a Beta prior, to be used in Automatic Relevance Detection

        It is exactly the same as a weight, except that it has a different prior, a Beta distribution prior between
        [0, 1].

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        parameter_settings = dict(sampling_prior=ARDBeta())

        super(ARD_Beta_Weight, self).__init__(name=name, value=value, lower_bound=lower_bound,
                                              upper_bound=upper_bound, parameter_kwargs=parameter_settings)


class ARD_Gaussian_Weight(Weight):

    def __init__(self, name='ARD_Gaussian_Weight', value=0.5, lower_bound=0.0, upper_bound=1.0):
        """A compartment weight with a Gaussian prior, to be used in Automatic Relevance Detection

        It is exactly the same as a weight, except that it has a different prior, a Gaussian prior with mean at zero
        and std given by a hyperparameter.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        parameter_settings = dict(sampling_prior=ARDGaussian())

        super(ARD_Gaussian_Weight, self).__init__(name=name, value=value, lower_bound=lower_bound,
                                                  upper_bound=upper_bound, parameter_kwargs=parameter_settings)



import inspect
import logging
from contextlib import contextmanager
from functools import reduce

import multiprocessing
import numpy as np
import pyopencl as cl
import numbers
from mot.cl_data_type import SimpleCLDataType
import pyopencl.array as cl_array
from scipy.special import jnp_zeros
import os

__author__ = 'Robbert Harms'
__date__ = "2014-05-13"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def dtype_to_ctype(dtype):
    """Get the CL type of the given numpy data type.

    Args:
        dtype (np.dtype): the numpy data type

    Returns:
        str: the CL type string for the corresponding type
    """
    from pyopencl.tools import dtype_to_ctype
    return dtype_to_ctype(dtype)


def ctype_to_dtype(cl_type, mot_float_type='float'):
    """Get the numpy dtype of the given cl_type string.

    Args:
        cl_type (str): the CL data type to match, for example 'float' or 'float4'.
        mot_float_type (str): the C name of the ``mot_float_type``. The dtype will be looked up recursively.

    Returns:
        dtype: the numpy datatype
    """
    data_type = SimpleCLDataType.from_string(cl_type)

    if data_type.is_vector_type:
        if data_type.raw_data_type.startswith('mot_float_type'):
            data_type = SimpleCLDataType.from_string(mot_float_type + str(data_type.vector_length))
        vector_type = data_type.raw_data_type + str(data_type.vector_length)
        return getattr(cl_array.vec, vector_type)
    else:
        if data_type.raw_data_type.startswith('mot_float_type'):
            data_type = SimpleCLDataType.from_string(mot_float_type)
        data_types = [
            ('char', np.int8),
            ('uchar', np.uint8),
            ('short', np.int16),
            ('ushort', np.uint16),
            ('int', np.int32),
            ('uint', np.uint32),
            ('long', np.int64),
            ('ulong', np.uint64),
            ('float', np.float32),
            ('double', np.float64),
        ]
        for ctype, dtype in data_types:
            if ctype == data_type.raw_data_type:
                return dtype


def get_bessel_roots(number_of_roots=30, np_data_type=np.float64):
    """These roots are used in some of the compartment models. It are the roots of the equation ``J'_1(x) = 0``.

    That is, where ``J_1`` is the first order Bessel function of the first kind.

    Args:
        number_of_root (int): The number of roots we want to calculate.
        np_data_type (np.data_type): the numpy data type

    Returns:
        ndarray: A vector with the indicated number of bessel roots (of the first order Bessel function
            of the first kind).
    """
    return jnp_zeros(1, number_of_roots).astype(np_data_type, copy=False, order='C')


def convert_data_to_dtype(data, data_type, mot_float_type='float'):
    """Convert the given input data to the correct numpy type.

    Args:
        value (ndarray): The value to convert to the correct numpy type
        data_type (str or mot.cl_data_type.CLDataType): the data type we need to convert the data to
        mot_float_type (str or mot.cl_data_type.CLDataType): the data type of the current ``mot_float_type``

    Returns:
        ndarray: the input data but then converted to the desired numpy data type
    """
    if isinstance(data_type, str):
        data_type = SimpleCLDataType.from_string(data_type)
    if isinstance(mot_float_type, str):
        mot_float_type = SimpleCLDataType.from_string(mot_float_type)

    scalar_dtype = ctype_to_dtype(data_type.raw_data_type, mot_float_type.raw_data_type)

    if isinstance(data, numbers.Number):
        data = scalar_dtype(data)

    if data_type.is_vector_type:
        shape = data.shape
        dtype = ctype_to_dtype(data_type, mot_float_type.raw_data_type)
        ve = np.zeros(shape[:-1], dtype=dtype)

        if len(shape) == 1:
            for vector_ind in range(shape[0]):
                ve[0][vector_ind] = data[vector_ind]
        elif len(shape) == 2:
            for i in range(data.shape[0]):
                for vector_ind in range(data.shape[1]):
                    ve[i][vector_ind] = data[i, vector_ind]
        elif len(shape) == 3:
            for i in range(data.shape[0]):
                for j in range(data.shape[1]):
                    for vector_ind in range(data.shape[2]):
                        ve[i, j][vector_ind] = data[i, j, vector_ind]

        return np.require(ve, requirements=['C', 'A', 'O'])
    return np.require(data, scalar_dtype, ['C', 'A', 'O'])


def device_type_from_string(cl_device_type_str):
    """Converts values like ``gpu`` to a pyopencl device type string.

    Supported values are: ``accelerator``, ``cpu``, ``custom``, ``gpu``. If ``all`` is given, None is returned.

    Args:
        cl_device_type_str (str): The string we want to convert to a device type.

    Returns:
        cl.device_type: the pyopencl device type.
    """
    cl_device_type_str = cl_device_type_str.upper()
    if hasattr(cl.device_type, cl_device_type_str):
        return getattr(cl.device_type, cl_device_type_str)
    return None


def device_supports_double(cl_device):
    """Check if the given CL device supports double

    Args:
        cl_device (pyopencl cl device): The device to check if it supports double.

    Returns:
        boolean: True if the given cl_device supports double, false otherwise.
    """
    return cl_device.get_info(cl.device_info.DOUBLE_FP_CONFIG) == 63


def get_float_type_def(double_precision):
    """Get the model floating point type definition.

    The MOT_INT_CMP_TYPE is meant for the select() function where you need a long in the case of double precision.

    Args:
        double_precision (boolean): if True we will use the double type for the mot_float_type type.
            Else, we will use the single precision float type for the mot_float_type type.

    Returns:
        str: defines the mot_float_type types, the epsilon and the MIN and MAX values.
    """
    if double_precision:
        return '''
            #if __OPENCL_VERSION__ <= CL_VERSION_1_1
                #pragma OPENCL EXTENSION cl_khr_fp64 : enable
            #endif

            #define mot_float_type double
            #define mot_float_type2 double2
            #define mot_float_type4 double4
            #define mot_float_type8 double8
            #define mot_float_type16 double16
            #define MOT_EPSILON DBL_EPSILON
            #define MOT_MIN DBL_MIN
            #define MOT_MAX DBL_MAX
            #define MOT_INT_CMP_TYPE long
        '''
    else:
        return '''
            #if __OPENCL_VERSION__ <= CL_VERSION_1_1
                #pragma OPENCL EXTENSION cl_khr_fp64 : enable
            #endif

            #define mot_float_type float
            #define mot_float_type2 float2
            #define mot_float_type4 float4
            #define mot_float_type8 float8
            #define mot_float_type16 float16
            #define MOT_EPSILON FLT_EPSILON
            #define MOT_MIN FLT_MIN
            #define MOT_MAX FLT_MAX
            #define MOT_INT_CMP_TYPE int
        '''


def topological_sort(data):
    """Topological sort the given dictionary structure.

    Args:
        data (dict); dictionary structure where the value is a list of dependencies for that given key.
            For example: ``{'a': (), 'b': ('a',)}``, where ``a`` depends on nothing and ``b`` depends on ``a``.

    Returns:
        tuple: the dependencies in constructor order
    """

    def check_self_dependencies(input_data):
        """Check if there are self dependencies within a node.

        Self dependencies are for example: ``{'a': ('a',)}``.

        Args:
            input_data (dict): the input data. Of a structure similar to {key: (list of values), ...}.

        Raises:
            ValueError: if there are indeed self dependencies
        """
        for k, v in input_data.items():
            if k in v:
                raise ValueError('Self-dependency, {} depends on itself.'.format(k))

    def prepare_input_data(input_data):
        """Prepares the input data by making sets of the dependencies. This automatically removes redundant items.

        Args:
            input_data (dict): the input data. Of a structure similar to {key: (list of values), ...}.

        Returns:
            dict: a copy of the input dict but with sets instead of lists for the dependencies.
        """
        return {k: set(v) for k, v in input_data.items()}

    def find_items_without_dependencies(input_data):
        """This searches the dependencies of all the items for items that have no dependencies.

        For example, suppose the input is: ``{'a': ('b',)}``, then ``a`` depends on ``b`` and ``b`` depends on nothing.
        This class returns ``(b,)`` in this example.

        Args:
            input_data (dict): the input data. Of a structure similar to {key: (list of values), ...}.

        Returns:
            list: the list of items without any dependency.
        """
        return list(reduce(set.union, input_data.values()) - set(input_data.keys()))

    def add_empty_dependencies(data):
        items_without_dependencies = find_items_without_dependencies(data)
        data.update({item: set() for item in items_without_dependencies})

    def get_sorted(input_data):
        data = input_data
        while True:
            ordered = set(item for item, dep in data.items() if len(dep) == 0)
            if not ordered:
                break
            yield ordered
            data = {item: (dep - ordered) for item, dep in data.items() if item not in ordered}

        if len(data) != 0:
            raise ValueError('Cyclic dependencies exist '
                             'among these items: {}'.format(', '.join(repr(x) for x in data.items())))

    check_self_dependencies(data)

    if not len(data):
        return []

    data_copy = prepare_input_data(data)
    add_empty_dependencies(data_copy)

    result = []
    for d in get_sorted(data_copy):
        try:
            d = sorted(d)
        except TypeError:
            d = list(d)

        result.extend(d)
    return result


def is_scalar(value):
    """Test if the given value is a scalar.

    This function also works with memory mapped array values, in contrast to the numpy is_scalar method.

    Args:
        value: the value to test for being a scalar value

    Returns:
        boolean: if the given value is a scalar or not
    """
    return np.isscalar(value) or (isinstance(value, np.ndarray) and (len(np.squeeze(value).shape) == 0))


def all_elements_equal(value):
    """Checks if all elements in the given value are equal to each other.

    If the input is a single value the result is trivial. If not, we compare all the values to see
    if they are exactly the same.

    Args:
        value (ndarray or number): a numpy array or a single number.

    Returns:
        bool: true if all elements are equal to each other, false otherwise
    """
    if is_scalar(value):
        return True
    return np.array(value == value.flatten()[0]).all()


def get_single_value(value):
    """Get a single value out of the given value.

    This is meant to be used after a call to :func:`all_elements_equal` that returned True. With this
    function we return a single number from the input value.

    Args:
        value (ndarray or number): a numpy array or a single number.

    Returns:
        number: a single number from the input

    Raises:
        ValueError: if not all elements are equal
    """
    if not all_elements_equal(value):
        raise ValueError('Not all values are equal to each other.')

    if is_scalar(value):
        return value
    return value.item(0)


@contextmanager
def all_logging_disabled(highest_level=logging.CRITICAL):
    """Disable all logging temporarily.

    A context manager that will prevent any logging messages triggered during the body from being processed.

    Args:
        highest_level: the maximum logging level that is being blocked
    """
    previous_level = logging.root.manager.disable
    logging.disable(highest_level)
    try:
        yield
    finally:
        logging.disable(previous_level)


class NamedCLFunction(object):

    def get_cl_code(self):
        """Return the CL function.

        Returns:
            str: the CL function
        """
        raise NotImplementedError()

    def get_cl_function_name(self):
        """Get the CL name of this function.

        Returns:
            str: the name of the function
        """
        raise NotImplementedError()


class SimpleNamedCLFunction(NamedCLFunction):

    def __init__(self, function, name):
        self._function = function
        self._name = name

    def get_cl_code(self):
        return self._function

    def get_cl_function_name(self):
        return self._name

    def __repr__(self):
        return self._function


def cartesian(arrays, out=None):
    """Generate a cartesian product of input arrays.

    Args:
        arrays (list of array-like): 1-D arrays to form the cartesian product of.
        out (ndarray): Array to place the cartesian product in.

    Returns:
        ndarray: 2-D array of shape (M, len(arrays)) containing cartesian products formed of input arrays.

    Examples:
        >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
        array([[1, 4, 6],
               [1, 4, 7],
               [1, 5, 6],
               [1, 5, 7],
               [2, 4, 6],
               [2, 4, 7],
               [2, 5, 6],
               [2, 5, 7],
               [3, 4, 6],
               [3, 4, 7],
               [3, 5, 6],
               [3, 5, 7]])
    """
    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    nmr_elements = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([nmr_elements, len(arrays)], dtype=dtype)

    m = nmr_elements // arrays[0].size
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m,1:])
        for j in range(1, arrays[0].size):
            out[j*m:(j+1)*m, 1:] = out[0:m, 1:]
    return out


def split_in_batches(nmr_elements, max_batch_size):
    """Split the total number of elements into batches of the specified maximum size.

    Examples::
        split_in_batches(30, 8) -> [(0, 8), (8, 15), (16, 23), (24, 29)]

        for batch_start, batch_end in split_in_batches(2000, 100):
            array[batch_start:batch_end]

    Yields:
        tuple: the start and end point of the next batch
    """
    offset = 0
    elements_left = nmr_elements
    while elements_left > 0:
        next_batch = (offset, offset + min(elements_left, max_batch_size))
        yield next_batch

        batch_size = min(elements_left, max_batch_size)
        elements_left -= batch_size
        offset += batch_size


def get_class_that_defined_method(method):
    """Get the class that defined the given method.

    This is taken from one of the answers of:
    ``https://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3/25959545``

    Args:
        method (func): a python function or method

    Returns:
        cls: the class that defined the given method
    """
    if inspect.ismethod(method):
        for cls in inspect.getmro(method.__self__.__class__):
            if cls.__dict__.get(method.__name__) is method:
                return cls
        method = method.__func__
    if inspect.isfunction(method):
        cls = getattr(inspect.getmodule(method),
                      method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
        if isinstance(cls, type):
            return cls
    return getattr(method, '__objclass__', None)


def hessian_to_covariance(hessian):
    """Calculate a covariance matrix from a Hessian by inverting the Hessian.

    Mathematically we can calculate the covariance matrix from the Hessian (the Hessian at the Maximum Likelihood
    Estimator), by a simple matrix inversion. However, round-off errors can make the Hessian singular, making an
    exact inverse impossible. This method uses an exact inverse if possible with a fall back on a pseudo inverse.

    Important: Before the matrix inversion it will set NaN's to 0. After the inversion we make the diagonal
    (representing the variances of each parameter) positive where needed by taking the absolute.

    Args:
        hessian (ndarray): a matrix of shape (n, p, p) where for n problems we have a matrix of shape (p, p) for
            p parameters and we take the inverse for every (p, p) matrix.

    Returns:
        ndarray: the covariance matrix calculated by inverting all the Hessians.
    """
    hessian = np.nan_to_num(hessian)

    covars = np.zeros_like(hessian)
    for roi_ind in range(hessian.shape[0]):
        try:
            covars[roi_ind] = np.linalg.inv(hessian[roi_ind])
        except np.linalg.linalg.LinAlgError:
            try:
                covars[roi_ind] = np.linalg.pinv(hessian[roi_ind])
            except np.linalg.linalg.LinAlgError:
                covars[roi_ind] = np.zeros(hessian[roi_ind].shape)

    diagonal_ind = np.arange(hessian.shape[1])
    covars[:, diagonal_ind, diagonal_ind] = np.abs(covars[:, diagonal_ind, diagonal_ind])

    return covars


def covariance_to_correlations(covariance):
    """Transform a covariance matrix into a correlations matrix.

    This can be seen as dividing a covariance matrix by the outer product of the diagonal.

    As post processing we replace the infinities and the NaNs with zeros and clip the result to [-1, 1].

    Args:
        covariance (ndarray): a matrix of shape (n, p, p) with for n problems the covariance matrix of shape (p, p).

    Returns:
        ndarray: the correlations matrix
    """
    diagonal_ind = np.arange(covariance.shape[1])
    diagonal_els = covariance[:, diagonal_ind, diagonal_ind]
    result = covariance / np.sqrt(diagonal_els[:, :, None] * diagonal_els[:, None, :])
    result[np.isinf(result)] = 0
    return np.clip(np.nan_to_num(result), -1, 1)


class KernelInputData(object):

    def set_mot_float_dtype(self, mot_float_dtype):
        """Set the numpy data type corresponding to the ``mot_float_type`` ctype.

        This is set by the kernel input data manager and is meant to update the state of the input data according
        to this setting.

        Args:
            mot_float_dtype (dtype): the numpy data type that is to correspond with the ``mot_float_type`` used in the
                kernels.
        """
        raise NotImplementedError()

    @property
    def is_scalar(self):
        """Check if the implemented input data is a scalar or not.

        Since scalars are loaded differently as buffers in the kernel, we have to check if this data is a scalar or not.

        Returns:
            boolean: if the implemented type should be loaded as a scalar or not
        """
        raise NotImplementedError()

    @property
    def read_data_back(self):
        """Check if this input data should be read back after kernel execution.

        Returns:
            boolean: if, after kernel launch, the data should be mapped from the compute device back to host memory.
        """
        raise NotImplementedError()

    def get_scalar_arg_dtype(self):
        """Get the numpy data type we should report in the kernel call for this input element.

        If we are inserting scalars in the kernel we need to provide the CL runtime with the correct data type
        of the function.

        Args:
            mot_float_dtype (dtype): the data type of the mot_float_type

        Returns:
            dtype: the numpy data type for this element, or None if this is not a scalar.
        """
        raise NotImplementedError()

    def get_data(self):
        """Get the underlying data.

        This should return the current state of the data object.

        Returns:
            the underlying data object, make sure this is of your desired data type. Can return None if
                this input data has no actual data.
        """
        raise NotImplementedError()

    def get_struct_declaration(self, name):
        """Get the declaration of this input data in the ``mot_data_struct`` object.

        Args:
            name (str): the name for this data, i.e. how it is represented in the kernel.

        Returns:
            str: the declaration of this input data in the kernel data struct
        """
        raise NotImplementedError()

    def get_kernel_argument_declaration(self, name):
        """Get the kernel argument declaration of this parameter.

        Args:
            name (str): the name for this data, i.e. how it is represented in the kernel.

        Returns:
            str: the kernel argument declaration
        """
        raise NotImplementedError()

    def get_struct_init_string(self, name, problem_id_substitute):
        """Create the structure initialization string.

        Args:
            name (str): the name of this data in the kernel
            problem_id_substitute (str): the substitute for the ``{problem_id}`` in the kernel data info elements.

        Returns:
            str: the instantiation string for the data struct of this input data
        """
        raise NotImplementedError()

    def get_kernel_inputs(self, cl_context, workgroup_size):
        """Get the kernel CL input object.

        Since the kernels follow the map/unmap paradigm make sure to use the ``USE_HOST_PTR`` when making
        writable data objects.

        Args:
            cl_context (pyopencl.Context): the CL context in which we are working.
            workgroup_size (int): the workgroup size the kernel will use

        Returns:
            a buffer, a local memory object, a scalars, etc., anything that can be loaded into the kernel.
        """
        raise NotImplementedError()

    def include_in_kernel_call(self):
        """Check if this data needs to be included in the kernel arguments.

        Returns:
            boolean: if the corresponding data needs to be included in the kernel call arguments. If set to False
                we typically expect the data to be inlined in the kernel. If set to True, we add the data to the
                kernel arguments.
        """
        raise NotImplementedError()


class KernelInputScalar(KernelInputData):

    def __init__(self, value, ctype=None):
        """A kernel input scalar.

        Args:
            value (number): the number to insert into the kernel as a scalar.
            ctype (str): the desired c-type for in use in the kernel, like ``int``, ``float`` or ``mot_float_type``.
                If None it is implied from the value.
        """
        self._value = np.array(value)
        self._ctype = ctype or dtype_to_ctype(self._value.dtype)
        self._mot_float_dtype = None

    def set_mot_float_dtype(self, mot_float_dtype):
        self._mot_float_dtype = mot_float_dtype

    @property
    def is_scalar(self):
        return True

    @property
    def read_data_back(self):
        return False

    def get_scalar_arg_dtype(self):
        if self._ctype.startswith('mot_float_type'):
            return self._mot_float_dtype
        return self._value.dtype

    def get_data(self):
        return self._value.astype(self.get_scalar_arg_dtype())

    def get_struct_declaration(self, name):
        return '{} {};'.format(self._ctype, name)

    def get_kernel_argument_declaration(self, name):
        return '{} {}'.format(self._ctype, name)

    def get_struct_init_string(self, name, problem_id_substitute):
        mot_dtype = SimpleCLDataType.from_string(self._ctype)
        if np.isinf(self._value):
            assignment = 'INFINITY'
        elif mot_dtype.is_vector_type:
            vector_length = mot_dtype.vector_length
            values = [str(val) for val in self._value[0]]
            if len(values) < vector_length:
                values.extend([str(0)] * (vector_length - len(values)))
            assignment = '(' + self._ctype + ')(' + ', '.join(values) + ')'
        else:
            assignment = str(self._value)
        return assignment

    def get_kernel_inputs(self, cl_context, workgroup_size):
        return self.get_data()

    def include_in_kernel_call(self):
        return False


class KernelInputLocalMemory(KernelInputData):

    def __init__(self, ctype, size_func=None):
        """Indicates that a local memory array of the indicated size must be loaded as kernel input data.

        Args:
            ctype (str): the desired c-type for this local memory object, like ``int``, ``float`` or ``mot_float_type``.
            size_func (Function): the function that can calculate the required local memory size (in number of bytes),
                given the workgroup size and numpy dtype in use by the kernel.
        """
        self._ctype = ctype
        self._size_func = size_func or (lambda workgroup_size, dtype: workgroup_size * np.dtype(dtype).itemsize)
        self._mot_float_dtype = None

    def set_mot_float_dtype(self, mot_float_dtype):
        self._mot_float_dtype = mot_float_dtype

    @property
    def is_scalar(self):
        return False

    @property
    def read_data_back(self):
        return False

    def get_scalar_arg_dtype(self):
        return None

    def get_data(self):
        return None

    def get_struct_declaration(self, name):
        return 'local {}* restrict {};'.format(self._ctype, name)

    def get_kernel_argument_declaration(self, name):
        return 'local {}* restrict {}'.format(self._ctype, name)

    def get_struct_init_string(self, name, problem_id_substitute):
        return name

    def get_kernel_inputs(self, cl_context, workgroup_size):
        return cl.LocalMemory(self._size_func(
            workgroup_size, ctype_to_dtype(self._ctype, dtype_to_ctype(self._mot_float_dtype))))

    def include_in_kernel_call(self):
        return True


class KernelInputArray(KernelInputData):

    def __init__(self, data, ctype=None, offset_str=None, is_writable=False, is_readable=True):
        """Loads the given array as a buffer into the kernel.

        By default, this will try to offset the data in the kernel by the stride of the first dimension multiplied
        with the problem id by the kernel. For example, if a (n, m) matrix is provided, this will offset the data
        by ``{problem_id} * m``.

        This class will adapt the data to match the ctype (if necessary) and it might copy the data as a consecutive
        array for direct memory access by the CL environment. Depending on the transformations, these values might
        or might not write to the same array as was provided. In short, do not trust your reference to the input data
        and always use :meth:`get_data` for retreiving return values.

        Args:
            data (ndarray): the data to load in the kernel
            ctype (str): the desired c-type for in use in the kernel, like ``int``, ``float`` or ``mot_float_type``.
                If None it is implied from the provided data.
            offset_str (str): the offset definition, can use ``{problem_id}`` for multiplication purposes. Set to 0
                for no offset.
            is_writable (boolean): if the data must be loaded writable or not, defaults to False
            is_readable (boolean): if this data must be made readable
        """
        self._requirements = ['C', 'A', 'O']
        if is_writable:
            self._requirements.append('W')

        self._data = np.require(data, requirements=self._requirements)
        if ctype and not ctype.startswith('mot_float_type'):
            self._data = convert_data_to_dtype(self._data, ctype)

        self._offset_str = offset_str
        self._is_writable = is_writable
        self._is_readable = is_readable
        self._ctype = ctype or dtype_to_ctype(data.dtype)
        self._mot_float_dtype = None
        self._backup_original_data = None

    def set_mot_float_dtype(self, mot_float_dtype):
        self._mot_float_dtype = mot_float_dtype

        if self._ctype.startswith('mot_float_type'):
            if self._backup_original_data is None:
                self._backup_original_data = np.copy(self._data)
            else:
                self._data = np.copy(self._backup_original_data)

            self._data = convert_data_to_dtype(self._data, self._ctype,
                                               mot_float_type=dtype_to_ctype(mot_float_dtype))
            self._data = np.require(self._data, requirements=self._requirements)

    def _get_offset_str(self, problem_id_substitute):
        if self._offset_str is None:
            offset_str = str(self._data.strides[0] // self._data.itemsize) + ' * {problem_id}'
        else:
            offset_str = str(self._offset_str)
        return offset_str.replace('{problem_id}', problem_id_substitute)

    @property
    def is_scalar(self):
        return False

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def read_data_back(self):
        return self._is_writable

    @property
    def is_readable(self):
        """If this kernel input data must be readable by the kernel.

        This is used in conjunction with :meth:`is_writable` when loading the data in the kernel.

        Returns:
            boolean: if this data must be made readable by the kernel function
        """
        return self._is_readable

    @property
    def is_writable(self):
        """Check if this kernel input data will write data back.

        This is used in conjunction with :meth:`is_readable` when loading the data in the kernel.

        If this returns true the kernel function must ensure that the data is loaded with at least write permissions.
        This flag will also ensure that the data will be read back from the device after kernel execution, overwriting
        the current data.

        Returns:
            boolean: if this data must be made writable and be read back after function execution.
        """
        return self._is_writable

    def get_struct_declaration(self, name):
        return 'global {}* restrict {};'.format(self._ctype, name)

    def get_kernel_argument_declaration(self, name):
        return 'global {}* restrict {}'.format(self._ctype, name)

    def get_struct_init_string(self, name, problem_id_substitute):
        return name + ' + ' + self._get_offset_str(problem_id_substitute)

    def get_scalar_arg_dtype(self):
        return None

    def get_data(self):
        return self._data

    def get_kernel_inputs(self, cl_context, workgroup_size):
        if self._is_writable:
            if self._is_readable:
                flags = cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR
            else:
                flags = cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR
        else:
            flags = cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR
        return cl.Buffer(cl_context, flags, hostbuf=self._data)

    def include_in_kernel_call(self):
        return True


class KernelInputAllocatedOutput(KernelInputData):

    def __init__(self, shape, ctype, offset_str=None, is_writable=True, is_readable=True):
        """Allocate an output buffer of the given shape.

        This is highly similar to :class:`~mot.utils.KernelInputArray` although it is writable by default.

        This is meant to quickly allocate a buffer large enough to hold the data requested. After running an OpenCL
        kernel you can get the written data using the method :meth:`get_data`.

        Args:
            shape (tuple): the shape of the output array
            offset_str (str): the offset definition, can use ``{problem_id}`` for multiplication purposes. Set to 0
                for no offset.
            is_writable (boolean): if the data must be loaded writable or not, defaults to True
            is_readable (boolean): if this data must be made readable, defaults to True
        """
        self._requirements = ['C', 'A', 'O']
        if is_writable:
            self._requirements.append('W')

        self._shape = shape
        self._data = None
        self._offset_str = offset_str
        self._is_writable = is_writable
        self._is_readable = is_readable
        self._ctype = ctype
        self._mot_float_dtype = None

    def set_mot_float_dtype(self, mot_float_dtype):
        dtype = ctype_to_dtype(self._ctype, mot_float_type=dtype_to_ctype(mot_float_dtype))
        self._mot_float_dtype = mot_float_dtype
        self._data = np.zeros(self._shape, dtype=dtype)
        self._data = np.require(self._data, requirements=self._requirements)

    def _get_offset_str(self, problem_id_substitute):
        if self._offset_str is None:
            offset_str = str(self._data.strides[0] // self._data.itemsize) + ' * {problem_id}'
        else:
            offset_str = str(self._offset_str)
        return offset_str.replace('{problem_id}', problem_id_substitute)

    @property
    def is_scalar(self):
        return False

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def read_data_back(self):
        return self._is_writable

    @property
    def is_readable(self):
        """If this kernel input data must be readable by the kernel.

        This is used in conjunction with :meth:`is_writable` when loading the data in the kernel.

        Returns:
            boolean: if this data must be made readable by the kernel function
        """
        return self._is_readable

    @property
    def is_writable(self):
        """Check if this kernel input data will write data back.

        This is used in conjunction with :meth:`is_readable` when loading the data in the kernel.

        If this returns true the kernel function must ensure that the data is loaded with at least write permissions.
        This flag will also ensure that the data will be read back from the device after kernel execution, overwriting
        the current data.

        Returns:
            boolean: if this data must be made writable and be read back after function execution.
        """
        return self._is_writable

    def get_struct_declaration(self, name):
        return 'global {}* restrict {};'.format(self._ctype, name)

    def get_kernel_argument_declaration(self, name):
        return 'global {}* restrict {}'.format(self._ctype, name)

    def get_struct_init_string(self, name, problem_id_substitute):
        return name + ' + ' + self._get_offset_str(problem_id_substitute)

    def get_scalar_arg_dtype(self):
        return None

    def get_data(self):
        return self._data

    def get_kernel_inputs(self, cl_context, workgroup_size):
        if self._is_writable:
            if self._is_readable:
                flags = cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR
            else:
                flags = cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR
        else:
            flags = cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR
        return cl.Buffer(cl_context, flags, hostbuf=self._data)

    def include_in_kernel_call(self):
        return True


class KernelInputDataManager(object):

    def __init__(self, kernel_input_dict, mot_float_dtype):
        """This class manages the transfer and definitions of the user input data into and from the kernel.

        Args:
            kernel_input_dict (dict[str: KernelInputData]): the kernel input data items by name
            mot_float_dtype (dtype): a numpy datatype indicating the data type we must use for inputs with ctype
                ``mot_float_type``.
        """
        self._kernel_input_dict = kernel_input_dict or []
        self._input_order = list(sorted(self._kernel_input_dict))
        self._data_duplicates = self.get_duplicate_items()
        self._mot_float_dtype = mot_float_dtype

        for input_data in kernel_input_dict.values():
            input_data.set_mot_float_dtype(mot_float_dtype)

    def get_duplicate_items(self):
        """Get a list of duplicate kernel inputs such that we don't load duplicate information twice.

        This will search for duplicate elements that are to be loaded in the kernel call, point to exactly the
        same memory and are defined to be write only.

        Returns:
            dict: a mapping from duplicate inputs (by name) to the only copy of the duplicate input object we load.
        """
        duplicates = {}

        for index in range(len(self._input_order)):
            key = self._input_order[index]
            kernel_input = self._kernel_input_dict[key]
            data = kernel_input.get_data()

            if not kernel_input.include_in_kernel_call() or kernel_input.read_data_back:
                continue

            if key not in duplicates:
                for other_index in range(index + 1, len(self._input_order)):
                    other_key = self._input_order[other_index]
                    other_kernel_input = self._kernel_input_dict[other_key]
                    other_data = other_kernel_input.get_data()

                    if not other_kernel_input.include_in_kernel_call() or other_kernel_input.read_data_back:
                        continue

                    if np.array_equal(data, other_data):
                        duplicates[other_key] = key

        return duplicates

    def get_struct_definition(self):
        """Return the structure definition of the mot_data_struct.

        This will use the sorted keys for looping through the kernel input items.

        Returns:
            str: the CL code for the structure definition
        """
        if not len(self._kernel_input_dict):
            return '''
                typedef struct{
                    constant void* place_holder;
                } mot_data_struct;
            '''

        definitions = []
        for name in self._input_order:
            kernel_input_data = self._kernel_input_dict[name]
            definitions.append(kernel_input_data.get_struct_declaration(name))

        return '''
            typedef struct{
                ''' + '\n'.join(definitions) + '''
            } mot_data_struct;
        '''

    def get_kernel_arguments(self):
        """Get the list of kernel arguments for loading the kernel data elements into the kernel.

        This will use the sorted keys for looping through the kernel input items.

        Returns:
            list of str: the list of parameter definitions
        """
        definitions = []
        for name in self._input_order:
            kernel_input_data = self._kernel_input_dict[name]
            if kernel_input_data.include_in_kernel_call() and name not in self._data_duplicates:
                definitions.append(kernel_input_data.get_kernel_argument_declaration(name))
        return definitions

    def get_struct_init_string(self, problem_id_substitute):
        """Create the structure initialization string.

        This will use the sorted keys for looping through the kernel input items.

        Args:
            problem_id_substitute (str): the substitute for the ``{problem_id}`` in the kernel data info elements.

        Returns:
            str: the instantiation string for the data struct
        """
        if not len(self._kernel_input_dict):
            return '{0}'

        definitions = []
        for name in self._input_order:

            if name in self._data_duplicates:
                name = self._data_duplicates[name]

            kernel_input_data = self._kernel_input_dict[name]
            definitions.append(kernel_input_data.get_struct_init_string(name, problem_id_substitute))

        return '{' + ', '.join(definitions) + '}'

    def get_kernel_inputs(self, cl_context, workgroup_size):
        """Get the kernel inputs to load.

        Args:
            cl_context (pyopencl.Context): the context in which we create the buffer
            workgroup_size (int): the workgroup size the kernel will use

        Returns:
            list of kernel input elements (buffers, local memory object, scalars, etc.)
        """
        kernel_inputs = []
        for name in self._input_order:
            data = self._kernel_input_dict[name]
            if data.include_in_kernel_call() and name not in self._data_duplicates:
                kernel_inputs.append(data.get_kernel_inputs(cl_context, workgroup_size))
        return kernel_inputs

    def get_scalar_arg_dtypes(self):
        """Get the location and types of the input scalars.

        Returns:
            list: for every kernel input element either None if the data is a buffer or the numpy data type if
                if is a scalar.
        """
        dtypes = []
        for ind, name in enumerate(self._input_order):
            data = self._kernel_input_dict[name]
            if data.include_in_kernel_call() and name not in self._data_duplicates:
                dtypes.append(data.get_scalar_arg_dtype())
        return dtypes

    def get_items_to_write_out(self):
        """Get the data name and buffer index of the items to write out after kernel execution.

        Returns:
            list: a list with (buffer index, name) tuples where the name refers to the name of the kernel input element
                and the index is the generated buffer index of that item.
        """
        items = []
        input_index = 0
        for name in self._input_order:
            if self._kernel_input_dict[name].read_data_back:
                items.append([input_index, name])
            if self._kernel_input_dict[name].include_in_kernel_call() and name not in self._data_duplicates:
                input_index += 1
        return items


def multiprocess_mapping(func, iterable):
    """Multiprocess mapping the given function on the given iterable.

    This only works in Linux and Mac systems since Windows has no forking capability. On Windows we fall back on
    single processing. Also, if we reach memory limits we fall back on single cpu processing.

    Args:
        func (func): the function to apply
        iterable (iterable): the iterable with the elements we want to apply the function on
    """
    if os.name == 'nt': # In Windows there is no fork.
        return list(map(func, iterable))
    try:
        p = multiprocessing.Pool()
        return_data = list(p.imap(func, iterable))
        p.close()
        p.join()
        return return_data
    except OSError:
        return list(map(func, iterable))




import re
from mot.cl_data_type import SimpleCLDataType
from mot.parsers.cl.CLDataType import CLDataTypeParser, CLDataTypeSemantics


class Semantics(CLDataTypeSemantics):

    def __init__(self):
        super(Semantics, self).__init__()
        self._raw_data_type = None
        self._is_pointer_type = False
        self._vector_length = None
        self._address_space_qualifier = None
        self._pre_asterisk_qualifiers = None
        self._post_asterisk_qualifiers = None

    def result(self, ast):
        return SimpleCLDataType(
            self._raw_data_type,
            is_pointer_type=self._is_pointer_type,
            vector_length=self._vector_length,
            address_space_qualifier=self._address_space_qualifier,
            pre_asterisk_qualifiers=self._pre_asterisk_qualifiers,
            post_asterisk_qualifiers=self._post_asterisk_qualifiers)

    def expr(self, ast):
        return ast

    def data_type(self, ast):
        return ast

    def is_pointer(self, ast):
        self._is_pointer_type = True
        return ast

    def scalar_data_type(self, ast):
        self._raw_data_type = ast
        return ast

    def vector_data_type(self, ast):
        match = re.match(r'(char|uchar|short|ushort|int|uint|long|ulong|float|double|half|mot_float_type)(\d+)', ast)
        self._raw_data_type = match.group(1)
        self._vector_length = match.group(2)
        return ast

    def user_data_type(self, ast):
        self._raw_data_type = ast
        return ast

    def address_space_qualifier(self, ast):
        self._address_space_qualifier = ast
        return ast

    def pre_asterisk_qualifiers(self, ast):
        self._pre_asterisk_qualifiers = ast
        return ast

    def post_asterisk_qualifiers(self, ast):
        self._post_asterisk_qualifiers = ast
        return ast


def parse(parameter_declaration):
    """Parse the parameter declaration into a CLDataType

    Args:
        parameter_declaration (str): the CL parameter declaration. Example: const float4* const test

    Returns:
        mot.cl_data_type.SimpleCLDataType: the CL data type for this parameter declaration
    """
    parser = CLDataTypeParser(parseinfo=False)
    return parser.parse(parameter_declaration, rule_name='result', semantics=Semantics())



import logging
from ...cl_routines.base import CLRoutine

__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractSampler(CLRoutine):

    def __init__(self, cl_environments=None, load_balancer=None, **kwargs):
        super(AbstractSampler, self).__init__(cl_environments=cl_environments, load_balancer=load_balancer, **kwargs)
        self._logger = logging.getLogger(__name__)

    def sample(self, model, init_params=None):
        """Sample the given model with the given codec using the given environments.

        Args:
            model (SampleModelInterface): the model to sample
            init_params (dict): a dictionary containing the results of a previous run, provides the starting point

        Returns:
            SamplingOutput: the sampling output object
        """
        raise NotImplementedError()


class SamplingOutput(object):

    def get_samples(self):
        """Get the matrix containing the sampling results.

        Returns:
            ndarray: the sampled parameter maps, a (d, p, n) array with for d problems and p parameters n samples.
        """
        raise NotImplementedError()

    def get_log_likelihoods(self):
        """Get per set of sampled parameters the log likelihood value associated with that set of parameters.

        Returns:
            ndarray: the log likelihood values, a (d, n) array with for d problems and n samples the log likelihood
                value.
        """
        raise NotImplementedError()

    def get_log_priors(self):
        """Get per set of sampled parameters the log prior value associated with that set of parameters.

        Returns:
            ndarray: the log prior values, a (d, n) array with for d problems and n samples the prior value.
        """
        raise NotImplementedError()


class SimpleSampleOutput(SamplingOutput):

    def __init__(self, samples, log_likelihoods, log_priors):
        """Simple storage container for the sampling output"""
        self._samples = samples
        self._log_likelihood = log_likelihoods
        self._log_prior = log_priors

    def get_samples(self):
        return self._samples

    def get_log_likelihoods(self):
        return self._log_likelihood

    def get_log_priors(self):
        return self._log_prior



from textwrap import dedent, indent

from mot.cl_parameter import CLFunctionParameter, SimpleCLFunctionParameter
from mot.cl_routines.mapping.cl_function_evaluator import CLFunctionEvaluator

__author__ = 'Robbert Harms'
__date__ = '2017-08-31'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class CLPrototype(object):
    """Prototype for a basic CL function."""

    def get_return_type(self):
        """Get the type (in CL naming) of the returned value from this function.

        Returns:
            str: The return type of this CL function. (Examples: double, int, double4, ...)
        """
        raise NotImplementedError()

    def get_cl_function_name(self):
        """Return the calling name of the implemented CL function

        Returns:
            str: The name of this CL function
        """
        raise NotImplementedError()

    def get_parameters(self):
        """Return the list of parameters from this CL function.

        Returns:
            list of CLFunctionParameter: list of the parameters in this model in the same order as in the CL function"""
        raise NotImplementedError()


class CLFunction(CLPrototype):
    """Interface for a basic CL function."""

    def get_cl_code(self):
        """Get the function code for this function and all its dependencies, with include guards.

        Returns:
            str: The CL code for inclusion in a kernel.
        """
        raise NotImplementedError()

    def get_cl_body(self):
        """Get the CL code of the body of this function.

        Returns:
            str: the CL code of this function body
        """
        raise NotImplementedError()

    def get_cl_extra(self):
        """Get the extra CL code outside the function body.

        Returns:
            str or None: the CL code outside the function body, can be None
        """
        raise NotImplementedError()

    def evaluate(self, inputs, double_precision=False, return_inputs=False):
        """Evaluate this function for each set of given parameters.

        Given a set of input parameters, this model will be evaluated for every parameter set.

        Args:
            inputs (dict[str: Union(ndarray, mot.utils.KernelInputData)]): for each parameter of the function
                the input data. Each of these input datasets must either be a scalar or be of equal length in the
                first dimension. The user can either input raw ndarrays or input KernelInputData objects.
                If an ndarray is given we will load it read/write by default.
            double_precision (boolean): if the function should be evaluated in double precision or not
            return_inputs (boolean): if we are interested in the values of the input arrays after evaluation.

        Returns:
            ndarray or tuple(ndarray, dict[str: ndarray]): we always return at least the return values of the function,
                which can be None if this function has a void return type. If ``return_inputs`` is set to True then
                we return a tuple with as first element the return value and as second element a dictionary mapping
                the output state of the parameters.
        """
        raise NotImplementedError()

    def get_dependencies(self):
        """Get the list of dependencies this function depends on.

        Returns:
            list[CLFunction]: the list of dependencies for this function.
        """
        raise NotImplementedError()


class SimpleCLPrototype(CLPrototype):

    def __init__(self, return_type, cl_function_name, parameter_list):
        """A simple implementation of a CL header.

        Args:
            return_type (str): the CL return type of the function
            cl_function_name (string): The name of the CL function
            parameter_list (list or tuple): This either contains instances of
                :class:`mot.cl_parameter.CLFunctionParameter` or contains tuples with arguments that
                can be used to construct a :class:`mot.cl_parameter.SimpleCLFunctionParameter`.
        """
        super(SimpleCLPrototype, self).__init__()
        self._return_type = return_type
        self._function_name = cl_function_name
        self._parameter_list = self._resolve_parameters(parameter_list)

    def get_cl_function_name(self):
        return self._function_name

    def get_return_type(self):
        return self._return_type

    def get_parameters(self):
        return self._parameter_list

    def _resolve_parameters(self, parameter_list):
        params = []
        for param in parameter_list:
            if isinstance(param, CLFunctionParameter):
                params.append(param)
            else:
                params.append(SimpleCLFunctionParameter(*param))
        return params

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return type(self) == type(other)

    def __ne__(self, other):
        return type(self) != type(other)


class SimpleCLFunction(CLFunction):

    def __init__(self, return_type, cl_function_name, parameter_list, cl_body, dependency_list=(), cl_extra=None):
        """A simple implementation of a CL function.

        Args:
            return_type (str): the CL return type of the function
            cl_function_name (string): The name of the CL function
            parameter_list (list or tuple): This either contains instances of
                :class:`mot.cl_parameter.CLFunctionParameter` or contains tuples with arguments that
                can be used to construct a :class:`mot.cl_parameter.SimpleCLFunctionParameter`.
            cl_body (str): the body of the CL code for this function.
            dependency_list (list or tuple of CLLibrary): The list of CL libraries this function depends on
            cl_extra (str): extra CL code for this function that does not warrant an own function
        """
        super(SimpleCLFunction, self).__init__()
        self._prototype = SimpleCLPrototype(return_type, cl_function_name, parameter_list)
        self._cl_body = cl_body
        self._dependency_list = dependency_list
        self._cl_extra = cl_extra

    def get_return_type(self):
        return self._prototype.get_return_type()

    def get_cl_function_name(self):
        return self._prototype.get_cl_function_name()

    def get_parameters(self):
        return self._prototype.get_parameters()

    def get_cl_code(self):
        cl_code = dedent('''
            {return_type} {cl_function_name}({parameters}){{
            {body}
            }}
        '''.format(return_type=self._prototype.get_return_type(),
                   cl_function_name=self._prototype.get_cl_function_name(),
                   parameters=', '.join(self._get_parameter_signatures()),
                   body=indent(dedent(self._cl_body), ' '*4*4)))

        return dedent('''
            {dependencies}
            #ifndef {inclusion_guard_name}
            #define {inclusion_guard_name}
            {cl_extra}
            {code}
            #endif // {inclusion_guard_name}
        '''.format(dependencies=indent(self._get_cl_dependency_code(), ' ' * 4 * 3),
                   inclusion_guard_name='INCLUDE_GUARD_{}'.format(self.get_cl_function_name()),
                   cl_extra=self._cl_extra if self._cl_extra is not None else '',
                   code=indent('\n' + cl_code + '\n', ' ' * 4 * 3)))

    def get_cl_body(self):
        return self._cl_body

    def get_cl_extra(self):
        return self._cl_extra

    def evaluate(self, inputs, double_precision=False, return_inputs=False):
        return CLFunctionEvaluator().evaluate(self, inputs, double_precision=double_precision,
                                              return_inputs=return_inputs)

    def get_dependencies(self):
        return self._dependency_list

    def _get_parameter_signatures(self):
        """Get the signature of the parameters for the CL function declaration.

        This should return the list of signatures of the parameters for use inside the function signature.

        Returns:
            list: the signatures of the parameters for the use in the CL code.
        """
        return ['{} {}'.format(p.data_type.get_declaration(), p.name.replace('.', '_'))
                for p in self._prototype.get_parameters()]

    def _get_cl_dependency_code(self):
        """Get the CL code for all the CL code for all the dependencies.

        Returns:
            str: The CL code with the actual code.
        """
        code = ''
        for d in self._dependency_list:
            code += d.get_cl_code() + "\n"
        return code

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return type(self) == type(other)

    def __ne__(self, other):
        return type(self) != type(other)



from copy import copy
import six
from mot.cl_data_type import SimpleCLDataType

__author__ = 'Robbert Harms'
__date__ = '2017-08-31'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class CLFunctionParameter(object):

    @property
    def data_type(self):
        """Get the CL data type of this parameter

        Returns:
            mot.cl_data_type.SimpleCLDataType: The CL data type.
        """
        raise NotImplementedError()

    @property
    def is_cl_vector_type(self):
        """Parse the data_type to see if this parameter holds a vector type (in CL)

        Returns:
            bool: True if the type of this function parameter is a CL vector type.

            CL vector types are recognized by an integer after the data type. For example: double4 is a
            CL vector type with 4 doubles.
        """
        raise NotImplementedError()

    def get_renamed(self, name):
        """Get a copy of the current parameter but then with a new name.

        Args:
            name (str): the new name for this parameter

        Returns:
            cls: a copy of the current type but with a new name
        """
        raise NotImplementedError()


class SimpleCLFunctionParameter(CLFunctionParameter):

    def __init__(self, data_type, name):
        """Creates a new function parameter for the CL functions.

        Args:
            data_type (mot.cl_data_type.SimpleCLDataType or str): the data type expected by this parameter
                If a string is given we will use ``SimpleCLDataType.from_string`` for translating the data_type.
            name (str): The name of this parameter

        Attributes:
            name (str): The name of this parameter
        """
        if isinstance(data_type, six.string_types):
            self._data_type = SimpleCLDataType.from_string(data_type)
        else:
            self._data_type = data_type

        self.name = name

    @property
    def data_type(self):
        return self._data_type

    @property
    def is_cl_vector_type(self):
        return self._data_type.is_vector_type

    def get_renamed(self, name):
        new_param = copy(self)
        new_param.name = name
        return new_param




from mot import configuration
from mot.configuration import get_compile_flags_to_disable_in_double_precision

__author__ = 'Robbert Harms'
__date__ = "2014-04-26"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CLRoutine(object):

    def __init__(self, cl_environments=None, load_balancer=None, compile_flags=None, **kwargs):
        """Base class for CL routines. Im

        Args:
            cl_environments (list of CLEnvironment): The list of CL environments using by this routine.
                If None is given we use the defaults in the current configuration.
            load_balancer (LoadBalancingStrategy): The load balancing strategy to be used by this routine.
                If None is given we use the defaults in the current configuration.
            compile_flags (dict): the list of compile flags to use during model fitting. As values use the
                flag name, as keys a boolean flag indicating if that one is active.
        """
        self._cl_environments = cl_environments
        self._load_balancer = load_balancer
        self.compile_flags = compile_flags

        if self._cl_environments is None:
            self._cl_environments = configuration.get_cl_environments()

        if self._load_balancer is None:
            self._load_balancer = configuration.get_load_balancer()

        if self.compile_flags is None:
            self.compile_flags = configuration.get_compile_flags(self.__class__.__name__)

    def get_cl_routine_kwargs(self):
        """Get a dictionary with the keyword arguments needed to create a similar CL routine instance.

        Returns:
            dict: a dictionary with the keyword arguments that a CLRoutine will take. This can be used to
                generate other CL routines with the same settings.
        """
        return dict(cl_environments=self.cl_environments,
                    load_balancer=self.load_balancer,
                    compile_flags=self.compile_flags)

    def set_compile_flag(self, compile_flag, enable):
        """Enable or disable the given compile flag.

        Args:
            compile_flag (str): the compile flag we want to enable or disable
            enable (boolean): if we enable (True) or disable (False) this compile flag
        """
        self.compile_flags.update({compile_flag: enable})

    def get_compile_flags_list(self, double_precision=True):
        """Get a list of the enabled compile flags.

        Args:
            double_precision (boolean): if this is set to True we remove some of the Flags that are only applicable
                when running in float mode. More specifically, this will set cl-single-precision-constant to False.
                Set this to False to disable this behaviour and use the flags as specified in the config.

        Returns:
            list: the list of enabled compile flags.
        """
        elements = [flag for flag, enabled in self.compile_flags.items() if enabled]

        if double_precision:
            elements_to_remove = get_compile_flags_to_disable_in_double_precision()
            elements = list(filter(lambda e: e not in elements_to_remove, elements))

        return elements

    @property
    def cl_environments(self):
        return self._cl_environments

    @cl_environments.setter
    def cl_environments(self, cl_environments):
        if cl_environments is not None:
            self._cl_environments = cl_environments

    @property
    def load_balancer(self):
        return self._load_balancer

    @load_balancer.setter
    def load_balancer(self, load_balancer):
        self._load_balancer = load_balancer

    def _create_workers(self, worker_generating_cb):
        """Create workers for all the CL environments in current use.

        Args:
            worker_generating_cb (python function): the callback function that we use to generate the
                worker for a specific CL environment. This should accept as single argument a CL environment and
                should return a Worker instance for use in CL computations.
        """
        cl_environments = self.load_balancer.get_used_cl_environments(self.cl_environments)
        return [worker_generating_cb(env) for env in cl_environments]



import warnings
import numpy as np
import pyopencl as cl
from mot.library_functions import Rand123
from ...cl_routines.sampling.base import AbstractSampler, SimpleSampleOutput
from ...load_balance_strategies import Worker
from ...utils import get_float_type_def, KernelInputDataManager, split_in_batches

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MetropolisHastings(AbstractSampler):

    def __init__(self, nmr_samples=None, burn_length=None,
                 sample_intervals=None, use_adaptive_proposals=True, **kwargs):
        """An CL implementation of Metropolis Hastings.

        This implementation uses a random walk single component updating strategy for the sampling.

        Args:
            nmr_samples (int): The length of the (returned) chain per voxel, defaults to 1
            burn_length (int): The length of the burn in (per voxel), these are extra samples,
                during burn-in we don't apply the thinning.
            sample_intervals (int): how many sample we wait before storing a new one.
                This will draw extra samples such that the total number of samples generated is
                ``chain_length * (sample_intervals + 1)`` and the number of samples stored is
                ``chain_length``. If set to zero we store every sample after the burn in.
            use_adaptive_proposals (boolean): if we use the adaptive proposals (set to True) or not (set to False).
        """
        super(MetropolisHastings, self).__init__(**kwargs)
        self._nmr_samples = nmr_samples or 1000
        self.burn_length = burn_length or 0
        self.sample_intervals = sample_intervals or 0
        self.use_adaptive_proposals = use_adaptive_proposals

        if self.burn_length is None:
            self.burn_length = 0

        if self.sample_intervals is None:
            self.sample_intervals = 0

        if self.burn_length < 0:
            raise ValueError('The burn length can not be smaller than 0, {} given'.format(self.burn_length))
        if self.sample_intervals < 0:
            raise ValueError('The sampling interval can not be smaller than 0, {} given'.format(self.sample_intervals))
        if self._nmr_samples < 1:
            raise ValueError('The number of samples to draw can '
                             'not be smaller than 1, {} given'.format(self._nmr_samples))

    @property
    def nmr_samples(self):
        return self._nmr_samples

    def sample(self, model, init_params=None):
        """Sample the given model with Metropolis Hastings

        Returns:
            MHSampleOutput: extension of the default output with some more data
        """
        mot_float_dtype = np.float32
        if model.double_precision:
            mot_float_dtype = np.float64

        self._do_initial_logging(model)

        if init_params is None:
            init_params = model.get_initial_parameters()

        nmr_params = init_params.shape[1]

        current_chain_position = np.require(init_params, mot_float_dtype, requirements=['C', 'A', 'O', 'W'])
        proposal_state = np.require(model.get_proposal_state(), mot_float_dtype, requirements=['C', 'A', 'O', 'W'])
        mh_state = _prepare_mh_state(model.get_metropolis_hastings_state(), mot_float_dtype)

        samples = np.zeros((model.get_nmr_problems(), nmr_params, self.nmr_samples),
                           dtype=mot_float_dtype, order='C')
        log_likelihoods = np.zeros((model.get_nmr_problems(), self.nmr_samples), dtype=mot_float_dtype, order='C')
        log_priors = np.zeros((model.get_nmr_problems(), self.nmr_samples), dtype=mot_float_dtype, order='C')

        def run(_samples, _log_likelihoods, _log_priors, _mh_state, nmr_samples, in_burnin=False):
            """Create the worker, process it, store the results in the given sample array and return a new mh state."""
            if in_burnin:
                sample_interval = 0
            else:
                sample_interval = self.sample_intervals

            workers = self._create_workers(lambda cl_environment: _MHWorker(
                cl_environment, self.get_compile_flags_list(model.double_precision), model, current_chain_position,
                _samples, _log_likelihoods, _log_priors, proposal_state, _mh_state, nmr_samples, in_burnin,
                sample_interval, self.use_adaptive_proposals, mot_float_dtype))
            self.load_balancer.process(workers, model.get_nmr_problems())
            return _mh_state.with_nmr_samples_drawn(_mh_state.nmr_samples_drawn + nmr_samples * (sample_interval + 1))

        self._logger.info('Starting sampling with method {0}'.format(self.__class__.__name__))

        if self.burn_length > 0:
            mh_state = run(samples, None, None, mh_state, self.burn_length, in_burnin=True)

        for batch_start, batch_end in split_in_batches(self.nmr_samples, 1000):
            batch_size = batch_end - batch_start

            samples_subset = np.zeros((model.get_nmr_problems(), nmr_params, batch_size),
                                      dtype=mot_float_dtype, order='C')
            ll_subset = np.zeros((model.get_nmr_problems(), batch_size), dtype=mot_float_dtype, order='C')
            lp_subset = np.zeros((model.get_nmr_problems(), batch_size), dtype=mot_float_dtype, order='C')

            mh_state = run(samples_subset, ll_subset, lp_subset, mh_state, batch_size)

            samples[..., batch_start:batch_end] = samples_subset
            log_likelihoods[..., batch_start:batch_end] = ll_subset
            log_priors[..., batch_start:batch_end] = lp_subset

        self._logger.info('Finished sampling')

        return MHSampleOutput(samples, log_likelihoods, log_priors, proposal_state, mh_state, current_chain_position)

    def _do_initial_logging(self, model):
        self._logger.info('Entered sampling routine.')
        self._logger.info('We will use a {} precision float type for the calculations.'.format(
            'double' if model.double_precision else 'single'))

        for env in self.load_balancer.get_used_cl_environments(self.cl_environments):
            self._logger.info('Using device \'{}\'.'.format(str(env)))

        self._logger.info('Using compile flags: {}'.format(self.get_compile_flags_list(model.double_precision)))

        sample_settings = dict(nmr_samples=self.nmr_samples,
                               burn_length=self.burn_length,
                               sample_intervals=self.sample_intervals,
                               use_adaptive_proposals=self.use_adaptive_proposals)
        self._logger.info('Sample settings: nmr_samples: {nmr_samples}, burn_length: {burn_length}, '
                          'sample_intervals: {sample_intervals}, '
                          'use_adaptive_proposals: {use_adaptive_proposals}. '.format(**sample_settings))

        samples_drawn = dict(samples_drawn=(self.burn_length + (self.sample_intervals + 1) * self.nmr_samples),
                             samples_returned=self.nmr_samples)
        self._logger.info('Total samples drawn: {samples_drawn}, total samples returned: '
                          '{samples_returned} (per problem).'.format(**samples_drawn))


class MHSampleOutput(SimpleSampleOutput):

    def __init__(self, samples, log_likelihoods, log_priors, proposal_state, mh_state, current_chain_position):
        """Simple storage container for the sampling output

        Args:
            samples (ndarray): an (d, p, n) matrix with d problems, p parameters and n samples
            log_likelihoods (ndarray): the log likelihood values, a (d, n) array with for d problems and n samples the
                log likelihood value.
            log_priors (ndarray): the log prior values, a (d, n) array with for d problems and n samples the
                prior value.
            proposal_state (ndarray): (d, p) matrix with for d problems and p parameters the proposal state
            mh_state (MHState): the current MH state
            current_chain_position (ndarray): (d, p) matrix with for d observations and p parameters the current
                chain position. If the samples are not empty the last element in the samples (``samples[..., -1]``)
                should equal this matrix.
        """
        super(MHSampleOutput, self).__init__(samples, log_likelihoods, log_priors)
        self._proposal_state = proposal_state
        self._mh_state = mh_state
        self._current_chain_position = current_chain_position

    def get_proposal_state(self):
        """Get the proposal state at the end of this sampling

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the proposal state
        """
        return self._proposal_state

    def get_mh_state(self):
        """Get the Metropolis Hastings state as it was at the end of this sampling run.

        Returns:
            MHState: the current MH state
        """
        return self._mh_state

    def get_current_chain_position(self):
        """Get the current chain position current_chain_position

        Returns:
            ndarray: (d, p) matrix with for d observations and p parameters the current
                chain position. If the samples are not empty the last element in the samples (``samples[..., -1]``)
                should equal this matrix.
        """
        return self._current_chain_position


class _MHWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, current_chain_position, samples,
                 log_likelihoods, log_priors, proposal_state, mh_state, nmr_samples, in_burnin,
                 sample_intervals, use_adaptive_proposals, mot_float_dtype):
        super(_MHWorker, self).__init__(cl_environment)

        self._model = model
        self._data_info = self._model.get_kernel_data()
        self._data_struct_manager = KernelInputDataManager(self._data_info, mot_float_dtype)
        self._current_chain_position = current_chain_position
        self._nmr_params = current_chain_position.shape[1]
        self._samples = samples
        self._log_likelihoods = log_likelihoods
        self._log_priors = log_priors

        self._proposal_state = proposal_state
        self._mh_state = mh_state
        self._mh_state_dict = self._get_mh_state_dict()

        self._nmr_samples = nmr_samples
        self._in_burnin = in_burnin
        self._sample_intervals = sample_intervals

        kernel_builder = _MCMCKernelBuilder(
            compile_flags, self._mh_state_dict, cl_environment, model,
            nmr_samples, not in_burnin, sample_intervals, use_adaptive_proposals, self._nmr_params, mot_float_dtype)

        self._kernel = kernel_builder.build()

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        workgroup_size = cl.Kernel(self._kernel, 'sample').get_work_group_info(
            cl.kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE, self._cl_environment.device)

        data_buffers, readout_items = self._get_buffers(workgroup_size)

        kernel_func = self._kernel.sample
        scalar_args = [None] * (5 + len(self._mh_state_dict))
        scalar_args.extend(self._data_struct_manager.get_scalar_arg_dtypes())

        if self._in_burnin:
            nmr_iterations = self._nmr_samples
        else:
            nmr_iterations = self._nmr_samples * (self._sample_intervals + 1)
            for item in [self._samples, self._log_likelihoods, self._log_priors]:
                buffer = cl.Buffer(self._cl_run_context.context,
                                   cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                   hostbuf=item)
                data_buffers.append(buffer)
                readout_items.append([buffer, item])
                scalar_args.append(None)

        kernel_func.set_scalar_arg_dtypes(scalar_args)

        kernel_func(
            self._cl_run_context.queue,
            (int(nmr_problems * workgroup_size),),
            (int(workgroup_size),),
            np.uint64(nmr_iterations),
            np.uint64(self._mh_state.nmr_samples_drawn),
            *data_buffers,
            global_offset=(range_start * workgroup_size,))

        for buffer, host_array in readout_items:
            self._enqueue_readout(buffer, host_array, range_start, range_end)

    def _get_buffers(self, workgroup_size):
        data_buffers = []
        readout_items = []

        current_chain_position_buffer = cl.Buffer(self._cl_run_context.context,
                                                  cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                                  hostbuf=self._current_chain_position)
        data_buffers.append(current_chain_position_buffer)
        readout_items.append([current_chain_position_buffer, self._current_chain_position])

        proposal_buffer = cl.Buffer(self._cl_run_context.context,
                                    cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                    hostbuf=self._proposal_state)
        data_buffers.append(proposal_buffer)
        readout_items.append([proposal_buffer, self._proposal_state])

        mcmc_state_buffers = {}
        for mcmc_state_element in sorted(self._mh_state_dict):
            host_array = self._mh_state_dict[mcmc_state_element]['data']

            buffer = cl.Buffer(self._cl_run_context.context,
                               cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                               hostbuf=host_array)
            mcmc_state_buffers[mcmc_state_element] = buffer

            data_buffers.append(buffer)
            readout_items.append([buffer, host_array])

        data_buffers.append(cl.LocalMemory(workgroup_size * np.dtype('double').itemsize))
        data_buffers.extend(self._data_struct_manager.get_kernel_inputs(self._cl_run_context.context, workgroup_size))

        return data_buffers, readout_items

    def _get_mh_state_dict(self):
        """Get a dictionary with the MH state kernel arrays"""
        state_dict = {
            'rng_state': {'data': self._mh_state.get_rng_state(),
                          'cl_type': 'uint'},
            'proposal_state_sampling_counter': {'data': self._mh_state.get_proposal_state_sampling_counter(),
                                                'cl_type': 'ulong'},
            'proposal_state_acceptance_counter': {'data': self._mh_state.get_proposal_state_acceptance_counter(),
                                                  'cl_type': 'ulong'}
        }

        if self._model.proposal_state_update_uses_variance():
            state_dict.update({
                'online_parameter_variance': {'data': self._mh_state.get_online_parameter_variance(),
                                              'cl_type': 'mot_float_type'},
                'online_parameter_variance_update_m2': {
                    'data': self._mh_state.get_online_parameter_variance_update_m2(), 'cl_type': 'mot_float_type'},
                'online_parameter_mean': {'data': self._mh_state.get_online_parameter_mean(),
                                          'cl_type': 'mot_float_type'},
            })

        return state_dict


class _MCMCKernelBuilder(object):

    def __init__(self, compile_flags, mh_state_dict, cl_environment, model,
                 nmr_samples, store_samples, sample_intervals, use_adaptive_proposals, nmr_params,
                 mot_float_dtype):
        self._cl_run_context = cl_environment.get_cl_context()
        self._compile_flags = compile_flags
        self._mh_state_dict = mh_state_dict
        self._cl_environment = cl_environment
        self._model = model
        self._data_info = self._model.get_kernel_data()
        self._data_struct_manager = KernelInputDataManager(self._data_info, mot_float_dtype)
        self._nmr_params = nmr_params
        self._nmr_samples = nmr_samples
        self._store_samples = store_samples
        self._sample_intervals = sample_intervals
        self._use_adaptive_proposals = use_adaptive_proposals
        self._update_parameter_variances = self._model.proposal_state_update_uses_variance()
        self._prior_func = self._model.get_log_prior_function(address_space_parameter_vector='local')
        self._proposal_func = self._model.get_proposal_function(address_space_proposal_state='global')
        self._proposal_logpdf_func = self._model.get_proposal_logpdf(address_space_proposal_state='global')
        self._proposal_state_update_func = self._model.get_proposal_state_update_function(address_space='global')

    def build(self):
        """Build the kernel according to the desired specifications.

        Returns:
            cl.Program: a compiled CL kernel
        """
        return self._compile_kernel(self._get_kernel_source())

    def _compile_kernel(self, kernel_source):
        from mot import configuration
        if configuration.should_ignore_kernel_compile_warnings():
            warnings.simplefilter("ignore")
        return cl.Program(self._cl_run_context.context, kernel_source).build(' '.join(self._compile_flags))

    def _get_kernel_source(self):

        kernel_param_names = [
            'ulong nmr_iterations',
            'ulong iteration_offset']

        kernel_param_names.extend([
            'global mot_float_type* restrict current_chain_position',
            'global mot_float_type* restrict global_proposal_state'])

        for mcmc_state_element in sorted(self._mh_state_dict):
            cl_type = self._mh_state_dict[mcmc_state_element]['cl_type']
            kernel_param_names.append('global {}* restrict global_{}'.format(cl_type, mcmc_state_element))

        kernel_param_names.append('local double* restrict log_likelihood_tmp')
        kernel_param_names.extend(self._data_struct_manager.get_kernel_arguments())

        if self._store_samples:
            kernel_param_names.append('global mot_float_type* restrict samples')
            kernel_param_names.append('global mot_float_type* restrict log_likelihoods')
            kernel_param_names.append('global mot_float_type* restrict log_priors')

        proposal_state_size = self._model.get_proposal_state().shape[1]

        kernel_source = '''
            #define NMR_INST_PER_PROBLEM ''' + str(self._model.get_nmr_inst_per_problem()) + '''
        '''
        kernel_source += get_float_type_def(self._model.double_precision)
        kernel_source += self._get_rng_functions()

        kernel_source += self._data_struct_manager.get_struct_definition()
        kernel_source += self._prior_func.get_cl_code()
        kernel_source += self._proposal_func.get_cl_code()

        if self._use_adaptive_proposals:
            kernel_source += self._proposal_state_update_func.get_cl_code()

        if not self._model.is_proposal_symmetric():
            kernel_source += self._proposal_logpdf_func.get_cl_code()

        kernel_source += self._get_log_likelihood_functions()
        kernel_source += self._get_state_update_functions()

        if self._update_parameter_variances:
            kernel_source += self._chain_statistics_update_function()

        kernel_source += '''
            void _sample(local mot_float_type* const x_local,
                         void* rng_data,
                         local double* const current_likelihood,
                         local mot_float_type* const current_prior,
                         mot_data_struct* data,
                         ulong nmr_iterations,
                         ulong iteration_offset,
                         global mot_float_type* const proposal_state,
                         global ulong* const sampling_counter,
                         global ulong* const acceptance_counter,
                         ''' + ('''global mot_float_type* parameter_mean,
                                   global mot_float_type* parameter_variance,
                                   global mot_float_type* parameter_variance_update_m2,'''
                                if self._update_parameter_variances else '') + '''
                         ''' + ('global mot_float_type* samples, '
                                'global mot_float_type* log_likelihoods, '
                                'global mot_float_type* log_priors, '
                                if self._store_samples else '') + '''
                         local double* log_likelihood_tmp){

                ulong i;
                uint j;
                ulong problem_ind = (ulong)(get_global_id(0) / get_local_size(0));
                bool is_first_work_item = get_local_id(0) == 0;

                for(i = 0; i < nmr_iterations; i++){
                '''
        if self._store_samples:
            kernel_source += '''
                    if(is_first_work_item){
                        if(i % ''' + str(self._sample_intervals + 1) + ''' == 0){

                            log_likelihoods[problem_ind * ''' + str(self._nmr_samples) + '''
                                            + (ulong)(i / ''' + str(self._sample_intervals + 1) + ''')
                                ] = *current_likelihood;
                            log_priors[problem_ind * ''' + str(self._nmr_samples) + '''
                                       + (ulong)(i / ''' + str(self._sample_intervals + 1) + ''')
                                ] = *current_prior;

                            for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                                samples[(ulong)(i / ''' + str(self._sample_intervals + 1) + ''') // remove the interval
                                        + j * ''' + str(self._nmr_samples) + '''  // parameter index
                                        + problem_ind * ''' + str(self._nmr_params * self._nmr_samples) + '''
                                    ] = x_local[j];
                            }
                        }
                    }
            '''
        if self._update_parameter_variances:
            kernel_source += '''
                    if(is_first_work_item){
                        for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                            _update_chain_statistics(i + iteration_offset,
                                x_local[j], parameter_mean + j, parameter_variance + j,
                                parameter_variance_update_m2 + j);
                        }
                    }
        '''
        kernel_source += '''
                    _update_state(x_local, rng_data, current_likelihood, current_prior,
                                  data, proposal_state, acceptance_counter,
                                  log_likelihood_tmp);

                    if(is_first_work_item){
                        for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                            sampling_counter[j]++;
                        }

                        ''' + (self._proposal_state_update_func.get_cl_function_name() +
                               '(proposal_state, sampling_counter, acceptance_counter' +
                                (', parameter_variance' if self._update_parameter_variances else '')
                                    + ');'
                               if self._use_adaptive_proposals else '') + '''
                    }
                }
            }
        '''
        kernel_source += '''
            __kernel void sample(
                ''' + ",\n".join(kernel_param_names) + '''
                ){

                ulong problem_ind = (ulong)(get_global_id(0) / get_local_size(0));

                mot_data_struct data = ''' + self._data_struct_manager.get_struct_init_string('problem_ind') + ''';

                rand123_data rand123_rng_data = _rng_data_from_array(global_rng_state);
                void* rng_data = (void*)&rand123_rng_data;

                local mot_float_type x_local[''' + str(self._nmr_params) + '''];

                local double current_likelihood;
                local mot_float_type current_prior;

                global mot_float_type* proposal_state =
                    global_proposal_state + problem_ind * ''' + str(proposal_state_size) + ''';
                global ulong* sampling_counter =
                    global_proposal_state_sampling_counter + problem_ind * ''' + str(self._nmr_params) + ''';
                global ulong* acceptance_counter =
                    global_proposal_state_acceptance_counter + problem_ind * ''' + str(self._nmr_params) + ''';
                '''
        if self._update_parameter_variances:
            kernel_source += '''
                global mot_float_type* parameter_mean =
                    global_online_parameter_mean + problem_ind * ''' + str(self._nmr_params) + ''';
                global mot_float_type* parameter_variance =
                    global_online_parameter_variance + problem_ind * ''' + str(self._nmr_params) + ''';
                global mot_float_type* parameter_variance_update_m2 =
                    global_online_parameter_variance_update_m2 + problem_ind * ''' + str(self._nmr_params) + ''';
            '''
        kernel_source += '''

                if(get_local_id(0) == 0){
                    for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        x_local[i] = current_chain_position[problem_ind * ''' + str(self._nmr_params) + ''' + i];
                    }

                    current_prior = ''' + self._prior_func.get_cl_function_name() + '''(&data, x_local);
                }
                barrier(CLK_LOCAL_MEM_FENCE);

                _fill_log_likelihood_tmp(&data, x_local, log_likelihood_tmp);
                _sum_log_likelihood_tmp(log_likelihood_tmp, &current_likelihood);

                _sample(x_local, rng_data, &current_likelihood, &current_prior, &data, nmr_iterations,
                        iteration_offset, proposal_state, sampling_counter, acceptance_counter,
                    ''' + ('parameter_mean, parameter_variance, parameter_variance_update_m2,'
                            if self._update_parameter_variances else '') + '''
                    ''' + ('samples, log_likelihoods, log_priors, ' if self._store_samples else '') + '''
                    log_likelihood_tmp);

                if(get_local_id(0) == 0){
                    for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        current_chain_position[problem_ind * ''' + str(self._nmr_params) + ''' + i] = x_local[i];
                    }

                    _rng_data_to_array(rand123_rng_data, global_rng_state);
                }
            }
        '''
        return kernel_source

    def _get_rng_functions(self):
        random_library = Rand123()
        kernel_source = random_library.get_cl_code()

        kernel_source += '''
            rand123_data _rng_data_from_array(global uint* rng_state){
                ulong problem_ind = (ulong)(get_global_id(0) / get_local_size(0));

                return rand123_initialize_data(
                    (uint[]){rng_state[0 + problem_ind * 6],
                             rng_state[1 + problem_ind * 6],
                             rng_state[2 + problem_ind * 6],
                             rng_state[3 + problem_ind * 6],
                             rng_state[4 + problem_ind * 6],
                             rng_state[5 + problem_ind * 6],
                             0, 0}); // last two states reserved for future use
            }

            void _rng_data_to_array(rand123_data data, global uint* rng_state){
                ulong problem_ind = (ulong)(get_global_id(0) / get_local_size(0));

                uint state[8]; // only using 6, other two reserved for future use
                rand123_data_to_array(data, state);

                for(uint i = 0; i < 6; i++){
                    rng_state[i + problem_ind * 6] = state[i];
                }
            }
        '''
        return kernel_source

    def _get_log_likelihood_functions(self):
        ll_func = self._model.get_log_likelihood_per_observation_function()

        kernel_source = ll_func.get_cl_code()
        kernel_source += '''
            void _fill_log_likelihood_tmp(mot_data_struct* data,
                                          local mot_float_type* const x_local,
                                          local double* log_likelihood_tmp){

                ulong observation_ind;
                ulong local_id = get_local_id(0);
                log_likelihood_tmp[local_id] = 0;
                uint workgroup_size = get_local_size(0);
                uint elements_for_workitem = ceil(NMR_INST_PER_PROBLEM / (mot_float_type)workgroup_size);

                if(workgroup_size * (elements_for_workitem - 1) + local_id >= NMR_INST_PER_PROBLEM){
                    elements_for_workitem -= 1;
                }

                mot_float_type x_private[''' + str(self._nmr_params) + '''];
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    x_private[i] = x_local[i];
                }

                for(uint i = 0; i < elements_for_workitem; i++){
                    observation_ind = i * workgroup_size + local_id;

                    log_likelihood_tmp[local_id] += ''' + ll_func.get_cl_function_name() + '''(
                        data, x_private, observation_ind);
                }

                barrier(CLK_LOCAL_MEM_FENCE);
            }

            void _sum_log_likelihood_tmp(local double* log_likelihood_tmp, local double* log_likelihood){
                *log_likelihood = 0;
                for(uint i = 0; i < get_local_size(0); i++){
                    *log_likelihood += log_likelihood_tmp[i];
                }
            }
        '''
        return kernel_source

    def _get_state_update_functions(self):
        kernel_source = '''
            void _update_state(local mot_float_type* const x_local,
                               void* rng_data,
                               local double* const current_likelihood,
                               local mot_float_type* const current_prior,
                               mot_data_struct* data,
                               global mot_float_type* const proposal_state,
                               global ulong * const acceptance_counter,
                               local double* log_likelihood_tmp){

                local float4 random_nmr;
                local mot_float_type new_prior;
                local double new_likelihood;
                local double bayesian_f;
                mot_float_type old_x;
                bool is_first_work_item = get_local_id(0) == 0;

                #pragma unroll 1
                for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                    if(is_first_work_item){
                        random_nmr = frand(rng_data);

                        old_x = x_local[k];
                        x_local[k] = ''' + self._proposal_func.get_cl_function_name() + \
                            '''(k, x_local[k], rng_data, proposal_state);

                        new_prior = ''' + self._prior_func.get_cl_function_name() + '''(data, x_local);
                    }
                    barrier(CLK_LOCAL_MEM_FENCE);

                    if(exp(new_prior) > 0){
                        _fill_log_likelihood_tmp(data, x_local, log_likelihood_tmp);

                        if(is_first_work_item){
                            _sum_log_likelihood_tmp(log_likelihood_tmp, &new_likelihood);

        '''
        if self._model.is_proposal_symmetric():
            kernel_source += '''
                            bayesian_f = exp((new_likelihood + new_prior) - (*current_likelihood + *current_prior));
                '''
        else:
            kernel_source += '''
                            mot_float_type x_to_prop = ''' + \
                             self._proposal_logpdf_func.get_cl_function_name() + '''(
                                k, old_x, x_local[k], proposal_state);
                            mot_float_type prop_to_x = ''' + \
                             self._proposal_logpdf_func.get_cl_function_name() + '''(
                                k, x_local[k], x_local[k], proposal_state);

                            bayesian_f = exp((new_likelihood + new_prior + x_to_prop) -
                                                (*current_likelihood + *current_prior + prop_to_x));
                '''
        kernel_source += '''
                            if(random_nmr.x < bayesian_f){
                                *current_likelihood = new_likelihood;
                                *current_prior = new_prior;
                                acceptance_counter[k]++;
                            }
                            else{
                                x_local[k] = old_x;
                            }
                        }
                    }
                    else{ // prior returned 0
                        if(is_first_work_item){
                            x_local[k] = old_x;
                        }
                    }
                }
            }
        '''
        return kernel_source

    def _chain_statistics_update_function(self):
        kernel_source = '''
            /** Online variance algorithm by Welford
             *  B. P. Welford (1962)."Note on a method for calculating corrected sums of squares
             *      and products". Technometrics 4(3):419-420.
             *
             * Also studied in:
             * Chan, Tony F.; Golub, Gene H.; LeVeque, Randall J. (1983).
             *      Algorithms for Computing the Sample Variance: Analysis and Recommendations.
             *      The American Statistician 37, 242-247. http://www.jstor.org/stable/2683386
             */
            void _update_chain_statistics(const ulong chain_count,
                                          const mot_float_type new_param_value,
                                          global mot_float_type* const parameter_mean,
                                          global mot_float_type* const parameter_variance,
                                          global mot_float_type* const parameter_variance_update_m2){

                mot_float_type previous_mean = *parameter_mean;
                *parameter_mean += (new_param_value - *parameter_mean) / (chain_count + 1);
                *parameter_variance_update_m2 += (new_param_value - previous_mean)
                                                    * (new_param_value - *parameter_mean);

                if(chain_count > 1){
                    *parameter_variance = *parameter_variance_update_m2 / (chain_count - 1);
                }
            }
        '''
        return kernel_source


class MHState(object):
    """The Metropolis Hastings state is used to initialize the state of the MH sampler.

    The state is stored at the end of every MH run and can be used to continue sampling again from the
    previous end point.
    """

    @property
    def nmr_samples_drawn(self):
        """Get the amount of samples already drawn, i.e. at what point in time is this state.

        Returns:
            uint64: the current number of samples already drawn before this state
        """
        raise NotImplementedError()

    def get_proposal_state_sampling_counter(self):
        """Get the current state of the sampling counter that can be used by the adaptive proposals.

        This value is per problem instance passed to the adaptive proposals which may reset the value.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current sampling counter,
                should be of a np.uint64 type.
        """
        raise NotImplementedError()

    def get_proposal_state_acceptance_counter(self):
        """Get the current state of the acceptance counter that can be used by the adaptive proposals.

        This value is per problem instance passed to the adaptive proposals which may reset the value.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current acceptance counter,
                should be of a np.uint64 type.
        """
        raise NotImplementedError()

    def get_online_parameter_variance(self):
        """Get the current state of the online parameter variance that can be used by the adaptive proposals.

        This value is updated while sampling and is passed as a constant to the adaptive proposals.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current parameter variance,
                should be of a np.float32 or np.float64 type (it will still be auto-converted to the
                current double type in the MCMC function).
        """
        raise NotImplementedError()

    def get_online_parameter_variance_update_m2(self):
        """A helper variable used in updating the online parameter variance.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current M2 state,
                should be of a np.float32 or np.float64 type (it will still be auto-converted to the
                current double type in the MCMC function).
        """
        raise NotImplementedError()

    def get_online_parameter_mean(self):
        """Get the current state of the online parameter mean, a helper variance in updating the variance.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current parameter mean,
                should be of a np.float32 or np.float64 type (it will still be auto-converted to the
                current double type in the MCMC function).
        """
        raise NotImplementedError()

    def get_rng_state(self):
        """Get the RNG state array for every problem instance.

        Returns:
            ndarray: a (d, \*) state array with for every d problem the state of size > 0
        """
        raise NotImplementedError()


class DefaultMHState(MHState):

    def __init__(self, nmr_problems, nmr_params, double_precision=False):
        """Creates a initial (default) MCMC state.

        Args:
            nmr_problems (int): the number of problems we are optimizing, used to create the default state.
            nmr_params (int): the number of parameters in the model, used to create the default state.
            double_precision (boolean): used when auto-creating some of the default state items.
        """
        self._nmr_problems = nmr_problems
        self._nmr_params = nmr_params
        self._double_precision = double_precision

        self._float_dtype = np.float32
        if double_precision:
            self._float_dtype = np.float64

    @property
    def nmr_samples_drawn(self):
        return 0

    def get_proposal_state_sampling_counter(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=np.uint64, order='C')

    def get_proposal_state_acceptance_counter(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=np.uint64, order='C')

    def get_online_parameter_variance(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=self._float_dtype, order='C')

    def get_online_parameter_variance_update_m2(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=self._float_dtype, order='C')

    def get_online_parameter_mean(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=self._float_dtype, order='C')

    def get_rng_state(self):
        dtype_info = np.iinfo(np.uint32)
        return np.random.uniform(low=dtype_info.min, high=dtype_info.max + 1,
                                 size=(self._nmr_problems, 6)).astype(np.uint32)


class SimpleMHState(MHState):
    def __init__(self, nmr_samples_drawn, proposal_state_sampling_counter, proposal_state_acceptance_counter,
                 online_parameter_variance, online_parameter_variance_update_m2,
                 online_parameter_mean,
                 rng_state):
        """A simple MCMC state containing provided items

        Args:
            nmr_samples_drawn (ndarray): the current number of samples already drawn to reach this state.
            proposal_state_sampling_counter (ndarray): a (d, p) array with for d problems and p parameters
                the current sampling counter.
            proposal_state_acceptance_counter (ndarray): a (d, p) array with for d problems and p parameters
                the current acceptance counter.
            online_parameter_variance (ndarray): a (d, p) array with for d problems and p parameters
                the current state of the online parameter variance
            online_parameter_variance_update_m2 (ndarray): a (d, p) array with for d problems and p parameters
                the current state of the online parameter variance update variable.
            online_parameter_mean (ndarray): a (d, p) array with for d problems and p parameters
                the current state of the online parameter mean
            rng_state (ndarray): a (d, \*) array with for d problems the rng state vector
        """
        self._nmr_samples_drawn = nmr_samples_drawn
        self._proposal_state_sampling_counter = proposal_state_sampling_counter
        self._proposal_state_acceptance_counter = proposal_state_acceptance_counter
        self._online_parameter_variance = online_parameter_variance
        self._online_parameter_variance_update_m2 = online_parameter_variance_update_m2
        self._online_parameter_mean = online_parameter_mean
        self._rng_state = rng_state

    @property
    def nmr_samples_drawn(self):
        return self._nmr_samples_drawn

    def with_nmr_samples_drawn(self, nmr_samples_drawn):
        """Recreate this object and set the number of samples drawn to the specified value."""
        return type(self)(
            nmr_samples_drawn,
            self.get_proposal_state_sampling_counter(),
            self.get_proposal_state_acceptance_counter(),
            self.get_online_parameter_variance(),
            self.get_online_parameter_variance_update_m2(),
            self.get_online_parameter_mean(),
            self.get_rng_state()
        )

    def get_proposal_state_sampling_counter(self):
        return self._proposal_state_sampling_counter

    def get_proposal_state_acceptance_counter(self):
        return self._proposal_state_acceptance_counter

    def get_online_parameter_variance(self):
        return self._online_parameter_variance

    def get_online_parameter_variance_update_m2(self):
        return self._online_parameter_variance_update_m2

    def get_online_parameter_mean(self):
        return self._online_parameter_mean

    def get_rng_state(self):
        return self._rng_state


def _prepare_mh_state(mh_state, float_dtype):
    """Return a new MH state in which all the state variables are sanitized to the correct data type.

    Args:
        mh_state (MHState): the MH state we wish to sanitize
        float_dtype (dtype): the numpy dtype for the floats, either np.float32 or np.float64

    Returns:
        SimpleMHState: MH state with the same data only then possibly sanitized
    """
    proposal_state_sampling_counter = np.require(np.copy(mh_state.get_proposal_state_sampling_counter()),
                                                 np.uint64,requirements=['C', 'A', 'O', 'W'])

    proposal_state_acceptance_counter = np.require(np.copy(mh_state.get_proposal_state_acceptance_counter()),
                                                   np.uint64, requirements=['C', 'A', 'O', 'W'])

    online_parameter_variance = np.require(np.copy(mh_state.get_online_parameter_variance()),
                                           float_dtype, requirements=['C', 'A', 'O', 'W'])

    online_parameter_variance_update_m2 = np.require(np.copy(mh_state.get_online_parameter_variance_update_m2()),
                                                     float_dtype, requirements=['C', 'A', 'O', 'W'])

    online_parameter_mean = np.require(np.copy(mh_state.get_online_parameter_mean()), float_dtype,
                                       requirements=['C', 'A', 'O', 'W'])

    rng_state = np.require(np.copy(mh_state.get_rng_state()), np.uint32, requirements=['C', 'A', 'O', 'W'])

    return SimpleMHState(mh_state.nmr_samples_drawn,
                         proposal_state_sampling_counter, proposal_state_acceptance_counter,
                         online_parameter_variance, online_parameter_variance_update_m2,
                         online_parameter_mean,
                         rng_state
                         )

